from dataclasses import InitVar, dataclass, field
from functools import lru_cache, partial
from itertools import count
from typing import Tuple, ClassVar, List

import numpy as np
from loguru import logger
from pyqtgraph import PlotItem, PlotWidget, TableWidget, mkColor, intColor, PlotDataItem, mkPen, mkBrush, colormap, \
    ScatterPlotItem, BarGraphItem
from qtpy.QtWidgets import QFormLayout, QWidget, QComboBox, QLabel, QVBoxLayout
from qtpy.QtCore import Qt, QSignalBlocker, Signal, QRectF

from tsuchinoko.graphics_items.mixins import ClickRequester, DomainROI, BetterButtons, LogScaleIntensity, \
    BetterAutoLUTRangeImageView, ViridisImageView, AspectRatioLock, YInvert
from tsuchinoko.graphs import Graph, Location, graph_signal_relay
from tsuchinoko.widgets.displays import Configuration
from tsuchinoko.widgets.graph_widgets import CloudWidget
import sklearn

from tsuchinoko.widgets.simple import DoubleSlider, ValueDoubleSlider


@dataclass(eq=False)
class Scatter(Graph):
    widget_class = PlotWidget

    def __init__(self, x_key: str, y_key: str, n_clusters: int, kmeans_kwargs: dict = None, name: str = 'Scatter'):
        self.x_key = x_key
        self.y_key = y_key
        self.n_clusters = n_clusters
        self.kmeans_kwargs = kmeans_kwargs
        super().__init__(name)

    def update(self, widget, data, update_slice: slice):
        with data.r_lock():
            x = data[self.x_key].copy()
            y = data[self.y_key].copy()

        widget.clear()

        kmeans = sklearn.cluster.KMeans(self.n_clusters, **self.kmeans_kwargs or {})
        kmeans.fit(np.dstack((x, y)))
        labels = kmeans.labels_
        for i in range(self.n_clusters):
            xi = x[labels == i]
            yi = y[labels == i]
            widget.addItem(ScatterPlotItem(x=xi, y=yi, brush=mkBrush(color=intColor(i, values=self.n_clusters))))


@dataclass(eq=False)
class Table(Graph):
    widget_class = TableWidget
    widget_kwargs: dict = field(default_factory=lambda: dict(sortable=False))
    data_keys: Tuple[str] = tuple()
    name = 'Table'

    def update(self, widget, data, update_slice: slice):
        # data = data[update_slice]

        with data.r_lock():
            x = data.positions.copy()
            v = data.variances.copy()
            y = data.scores.copy()

            extra_fields = {data_key: data[data_key].copy() for data_key in self.data_keys}

        lengths = len(v), len(x), len(y), *map(len, extra_fields.values())
        min_length = min(lengths)
        if not np.all(np.array(lengths) == min_length):
            logger.warning(f'Ragged arrays passed to cloud item with lengths (v, x, y): {lengths}')
            x = x[:min_length]
            y = y[:min_length]
            v = v[:min_length]
            extra_fields = {k: v[:min_length] for k, v in extra_fields.items()}

        values = [x, y, v, *extra_fields.values()]

        names = ['Position', 'Value', 'Variance'] + list(extra_fields.keys())

        rows = range(update_slice.start, len(x))
        table = [{name: value[i] for name, value in zip(names, values)} for i in rows]

        if update_slice.start == 0:
            widget.setData(table)
        else:
            for row, table_row in zip(rows, table):
                widget.setRow(row, list(table_row.values()))


class ImageViewBlend(YInvert,
                     ClickRequester,
                     BetterButtons,
                     LogScaleIntensity,
                     AspectRatioLock,
                     BetterAutoLUTRangeImageView,
                     ViridisImageView):
    pass


class ImageViewBlendROI(DomainROI, ImageViewBlend):
    pass


# TODO: add option for transforming into parameter space or not

@dataclass(eq=False)
class Image(Graph):
    widget_class = ImageViewBlend
    data_key: ClassVar[str] = None
    accumulates: ClassVar[bool] = False
    transform_to_parameter_space = True
    widget_kwargs: dict = field(default_factory=lambda: dict(invert_y=False))

    def __post_init__(self):
        if not self.name and self.data_key:
            self.name = self.data_key

    def update(self, widget, data, update_slice: slice):
        with data.r_lock():
            try:
                v = data[self.data_key].copy()
            except ValueError:
                if getattr(self, '_has_value_errors', False):
                    pass
                    # logger.warning(f'The {self.name} graph hasn\'t received data more than once. This is not normal.')
                else:
                    logger.info(f'The {self.name} graph hasn\'t received data once. This is normal for graphs computed by the adaptive engine.')
                self._has_value_errors = True
                return

        if self.accumulates:
            raise NotImplemented('Accumulation in Image graphs not implemented yet')
        else:
            if getattr(v, 'ndim', None) in [2, 3]:
                bounds = [tuple(Configuration().parameter.child('bounds')[f'axis_{i}_{limit}']
                                for limit in ['min', 'max'])
                          for i in range(2)]
                axes = None
                if v.ndim == 2:
                    axes = {'x': 0, 'y': 1}
                elif v.ndim == 3:
                    # if shape is 3d and #3 is 3, use color, otherwise use t for dimensions
                    if v.shape[2] == 3:
                        axes = {'x': 0, 'y': 1, 'c': 2}
                    else:
                        axes = {'t': 2, 'x': 0, 'y': 1}

                kwargs = {}
                if self.transform_to_parameter_space:
                    kwargs['pos'] = (bounds[0][0], bounds[1][0])
                    kwargs['scale'] = ((bounds[0][1] - bounds[0][0]) / v.shape[0], (bounds[1][1] - bounds[1][0]) / v.shape[1])

                widget.setImage(np.array(v),
                                autoRange=widget.imageItem.image is None,
                                autoLevels=False, #widget.imageItem.image is None,
                                autoHistogramRange=widget.imageItem.image is None,
                                axes=axes,
                                **kwargs)


@dataclass(eq=False)
class Cloud(Graph):
    data_key: ClassVar[str] = None
    accumulates: ClassVar[bool] = True
    widget_class = type('CloudBlend', (CloudWidget, DomainROI), {})
    widget_args: Tuple = tuple()

    def __post_init__(self):
        self.widget_args = (self.data_key, self.accumulates)

    def update(self, widget, data, update_slice: slice):
        widget.update_data(data, update_slice)


class PlotGraphWidget(PlotWidget):
    def __init__(self, *args, label_key=None, **kwargs):
        super().__init__(*args, **kwargs)
        if label_key:
            self.getPlotItem().addLegend()


@dataclass(eq=False)
class Bar(Graph):
    data_key: ClassVar[str] = None
    widget_class = PlotGraphWidget
    label_key: InitVar[str] = None

    def __post_init__(self, label_key):
        self.widget_kwargs['label_key'] = label_key

    def update(self, widget, data, update_slice: slice):
        if not hasattr(self, 'bgi'):
            self.bgi = BarGraphItem(x0=[], x1=[], height=[], pen='w', brush=(0, 0, 255, 150))
            widget.addItem(self.bgi)

        with data.r_lock():
            if self.data_key in data:
                y, x = data[self.data_key].copy()
            else:
                logger.warning(f'A graph could not find the required key: {self.data_key}.')

        self.bgi.setOpts(x0=x[:-1], x1=x[1:], height=y)




@dataclass(eq=False)
class Plot(Graph):
    data_key: ClassVar[str] = None
    widget_class = PlotGraphWidget
    label_key: InitVar[str] = None
    accumulates: bool = False

    def __post_init__(self, label_key):
        self.widget_kwargs['label_key'] = label_key

    def update(self, widget, data, update_slice: slice):
        with data.r_lock():
            v = data[self.data_key].copy()
        if self.accumulates:
            widget.plot(np.asarray(v), clear=True, label=self.label_key)
        else:
            widget.plot(np.asarray(v), clear=True, label=self.label_key)


@dataclass(eq=False)
class MultiPlot(Plot):
    pen_key:str = None
    stack_plots: ClassVar[bool] = True

    @staticmethod
    def get_color(i, count):
        if count < 9:
            color = mkColor(i)
        else:
            color = intColor(i, hues=count, minHue=180, maxHue=300)
        return color

    def colorize(self, widget, data):
        plot_data_items = list(filter(lambda item: isinstance(item, PlotDataItem), widget.getPlotItem().items))
        count = len(plot_data_items)

        for i, item in enumerate(plot_data_items):
            if isinstance(item, PlotDataItem):
                color = self.get_color(i, count)
                item.setPen(color)
                item.setSymbolBrush(color)
                item.setSymbolPen('w')

    def update(self, widget, data: 'Data', update_slice: slice):
        if update_slice.start == 0 or not self.stack_plots:
            widget.getPlotItem().clear()

        with data.r_lock():
            try:
                v = data[self.data_key].copy()
            except ValueError:
                if getattr(self, '_has_value_errors', False):
                    logger.warning(f'The {self.name} graph hasn\'t received data more than once. This is not normal.')
                else:
                    logger.info(
                        f'The {self.name} graph hasn\'t received data once. This is normal for graphs computed by the adaptive engine.')
                self._has_value_errors = True
                return
            labels = data[self.label_key].copy()
            if self.pen_key is not None:
                pens = data[self.pen_key].copy()

        if self.stack_plots:
            plots = zip(count(update_slice.start), labels[update_slice], v[update_slice])
        else:
            plots = zip(count(len(labels)), labels, v)

        for i, label, plot_data in plots:
            kwargs = {}
            if self.pen_key is not None:
                kwargs['pen'] = mkPen(pens[i])
            widget.plot(np.array(plot_data), name=label, **kwargs)

        if self.pen_key is None:
            self.colorize(widget, data)


@dataclass(eq=False)
class DynamicColorMultiPlot(MultiPlot):
    def __init__(self, color_scalar_key, *args, colormap_name='CET-L17', **kwargs):
        super().__init__(*args, **kwargs)
        self.colormap = colormap.get(colormap_name)
        self.color_scalar_key = color_scalar_key
        self.item_colors = []

    def update(self, widget, data: 'Data', update_slice: slice):
        with data.r_lock():
            c = data[self.color_scalar_key].copy()
        c_min = np.min(c)
        c_max = np.max(c)
        scaled_c = np.interp(c, (c_min, c_max), (0, 1))
        self.item_colors = list(map(self.colormap.map, scaled_c))

        super().update(widget, data, update_slice)

    def get_color(self, i, count):
        return self.item_colors[i]


@dataclass(eq=False)
class Variance(Cloud):
    data_key = 'variances'
    name = 'Variance'

    def compute(self, data, engine):
        pass  # This is free


@dataclass(eq=False)
class Score(Cloud):
    data_key = 'scores'
    name = 'Score'

    def compute(self, data, engine):
        pass  # This is free


@dataclass(eq=False)
class GPCamPosteriorCovariance(Image):
    shape = (50, 50)
    data_key = 'Posterior Covariance'
    widget_kwargs: dict = field(default_factory=lambda: dict(invert_y=True))
    transform_to_parameter_space: ClassVar[bool] = False

    def compute(self, data, engine: 'GPCamInProcessEngine'):
        with data.r_lock():  # quickly grab positions within lock before passing to optimizer
            positions = np.asarray(data.positions.copy())

        # if multi-task, extend the grid_positions to include the task dimension
        if hasattr(engine, 'output_number'):
            positions = np.vstack([np.hstack([positions, np.full((positions.shape[0], 1), i)]) for i in range(engine.output_number)])

        # compute posterior covariance without lock
        result_dict = engine.optimizer.posterior_covariance(positions)

        # assign to data object with lock
        with data.w_lock():
            data.states[self.data_key] = result_dict['S']


@dataclass(eq=False)
class GPCamAcquisitionFunction(Image):
    compute_with = Location.AdaptiveEngine
    shape:tuple = (50, 50)
    data_key = 'Acquisition Function'
    widget_class = ImageViewBlendROI

    def compute(self, data, engine: 'GPCAMInProcessEngine'):
        from tsuchinoko.adaptive.gpCAM_in_process import gpcam_acquisition_functions  # avoid circular import

        bounds = tuple(tuple(engine.parameters[('bounds', f'axis_{i}_{edge}')]
                             for edge in ['min', 'max'])
                       for i in range(engine.dimensionality))

        grid_positions = image_grid(bounds, self.shape)

        # check if acquisition function is accessible
        if engine.parameters['acquisition_function'] not in gpcam_acquisition_functions:
            logger.exception(ValueError('The selected acquisition_function is not available for display.'))
            return

        extra_kwargs={}
        output_num = getattr(engine.optimizer.gp, 'output_num', 1)
        if output_num > 1:
            extra_kwargs['x_out'] = np.arange(output_num)

        # calculate acquisition function
        acquisition_function_value = engine.optimizer.evaluate_acquisition_function(grid_positions,
                                                                                    acquisition_function=
                                                                                    gpcam_acquisition_functions[
                                                                                        engine.parameters[
                                                                                            'acquisition_function']],
                                                                                    origin=engine.last_position,
                                                                                    **extra_kwargs)

        try:
            acquisition_function_value = acquisition_function_value.reshape(*self.shape)
        except (ValueError, AttributeError):
            acquisition_function_value = np.array([[0]])

        # assign to data object with lock
        with data.w_lock():
            data.states[self.data_key] = acquisition_function_value


@dataclass(eq=False)
class GPCamPosteriorMean(Image):
    compute_with = Location.AdaptiveEngine
    shape:tuple = (50, 50)
    data_key = 'Posterior Mean'
    widget_class = ImageViewBlendROI

    def compute(self, data, engine: 'GPCAMInProcessEngine'):
        bounds = tuple(tuple(engine.parameters[('bounds', f'axis_{i}_{edge}')]
                   for edge in ['min', 'max'])
                  for i in range(engine.dimensionality))

        grid_positions = image_grid(bounds, self.shape)
        shape = self.shape

        # For GITOMO; needs to be combined with below
        # if multi-task, extend the grid_positions to include the task dimension
        # if hasattr(engine, 'output_number'):
        #     grid_positions = np.vstack([np.hstack([grid_positions, np.full((grid_positions.shape[0], 1), i)]) for i in range(engine.output_number)])
        #     shape = (*self.shape, engine.output_number)

        extra_kwargs = dict()
        if hasattr(engine.optimizer.gp, 'output_num'):
            extra_kwargs['x_out'] = np.arange(engine.optimizer.gp.output_num)
            shape = (*shape, engine.optimizer.gp.output_num)

        # calculate acquisition function
        posterior_mean_value = engine.optimizer.posterior_mean(grid_positions, **extra_kwargs)['m(x)'].reshape(*shape)

        # assign to data object with lock
        with data.w_lock():
            data.states['Posterior Mean'] = posterior_mean_value


@lru_cache(maxsize=10)
def image_grid(bounds, shape):
    return np.asarray(np.meshgrid(*(np.linspace(bound[0], bound[1]-1, num=bins, endpoint=True)
                                    for bins, bound in zip(shape, bounds)))).T.reshape(-1, 2)


class SliceImageWidget(QWidget):
    sigSliceChanged = Signal(int, float)  # axis, value
    sigImageAxesChanged = Signal(int, int)  # x and y image axes

    def __init__(self, dimensions: int, bounds:List[List[float]], invert_y=False):
        super().__init__()
        dim_indices = list(range(1, dimensions + 1))
        self.x_dimension_selector = QComboBox()
        self.y_dimension_selector = QComboBox()
        self.x_dimension_selector.addItems(map(str, dim_indices))
        self.y_dimension_selector.addItems(map(str, dim_indices))
        self.x_dimension_selector.setCurrentIndex(0)
        self.y_dimension_selector.setCurrentIndex(1)
        self.x_dimension_selector.currentIndexChanged.connect(partial(self.dimension_selected, axis='x'))
        self.y_dimension_selector.currentIndexChanged.connect(partial(self.dimension_selected, axis='y'))

        self.image_view = ImageViewBlend(invert_y=invert_y)

        self.dims_layout = QFormLayout()
        self.setLayout(QVBoxLayout())
        self.layout().addWidget(self.image_view)
        self.layout().addLayout(self.dims_layout)

        self.dims_layout.addRow("X dimension:", self.x_dimension_selector)
        self.dims_layout.addRow("Y dimension:", self.y_dimension_selector)
        self.dimension_sliders = []
        for i in range(dimensions):
            slider = ValueDoubleSlider(Qt.Horizontal)
            slider.sigFloatValueChanged.connect(partial(self.sigSliceChanged.emit, i))
            slider.setMinimum(bounds[i][0])
            slider.setMaximum(bounds[i][1])
            slider.setInterval((bounds[i][1] - bounds[i][0]) / 100)
            self.dimension_sliders.append(slider)
        for i, slider in enumerate(self.dimension_sliders[2:]):
            label = QLabel(f'Dimension {i + 3} slice:')
            self.dims_layout.addRow(label, slider)

    def dimension_selected(self, index, axis):
        dim_indices = set(range(self.x_dimension_selector.count()))
        dim_indices.remove(index)
        if axis == 'x':
            if self.y_dimension_selector.currentIndex() == index:
                blocker = QSignalBlocker(self.y_dimension_selector)
                self.y_dimension_selector.setCurrentIndex(dim_indices.pop())
            else:
                dim_indices.remove(self.y_dimension_selector.currentIndex())
        if axis == 'y':
            if self.x_dimension_selector.currentIndex() == index:
                blocker = QSignalBlocker(self.x_dimension_selector)
                self.x_dimension_selector.setCurrentIndex(dim_indices.pop())
            else:
                dim_indices.remove(self.x_dimension_selector.currentIndex())

        # clear layout
        for i in reversed(range(self.dims_layout.count())):
            self.dims_layout.itemAt(i).widget().setParent(None)

        self.dims_layout.addRow("X dimension:", self.x_dimension_selector)
        self.dims_layout.addRow("Y dimension:", self.y_dimension_selector)
        for i in dim_indices:
            label = QLabel(f'Dimension {i + 1} slice:')
            self.dims_layout.addRow(label, self.dimension_sliders[i])

        self.sigImageAxesChanged.emit(self.x_dimension_selector.currentIndex(), self.y_dimension_selector.currentIndex())


@dataclass(eq=False)
class SliceImageGraph(Graph):
    dimensions: int = 3
    widget_class = SliceImageWidget
    bounds: List[List[float]] = field(default=None)
    data_key: ClassVar[str] = None
    invert_y:bool = False
    accumulates: ClassVar[bool] = False

    def __post_init__(self):
        self.dimension_sliders = []
        self.slices = [dim[0] for i, dim in enumerate(self.bounds)]
        self.image_axes = {'x': 0, 'y': 1}
        self.widget_kwargs['dimensions'] = self.dimensions
        self.widget_kwargs['bounds'] = self.bounds
        if not self.name and self.data_key:
            self.name = self.data_key

    def make_widget(self):
        widget = super().make_widget()
        widget.sigSliceChanged.connect(self.set_slice)
        widget.sigImageAxesChanged.connect(self.set_image_axes)
        return widget

    def set_slice(self, i, v):
        self.slices[i] = v
        graph_signal_relay.sigPush.emit(self)

    def set_image_axes(self, x, y):
        self.image_axes['x'] = x
        self.image_axes['y'] = y
        graph_signal_relay.sigPush.emit(self)

    def update(self, widget, data: 'Data', *args):
        Image.update(self, widget.image_view, data, *args)


@lru_cache(maxsize=10)
def slice_grid(bounds, shape, image_axes, slices):
    ticks = [*(np.linspace(*bounds[i],
                           num=shape[0] if image_axes[0] == i else shape[1])
               if i in image_axes else
               slices[i] for i in range(len(bounds)))]
    return np.asarray(np.meshgrid(*ticks)).T.reshape(-1, len(bounds))


class HighDimensionalityGPCamPosteriorMean(SliceImageGraph):
    compute_with = Location.AdaptiveEngine
    shape = (50, 50)
    data_key = 'Posterior Mean'

    def compute(self, data, engine: 'GPCAMInProcessEngine'):
        bounds = [tuple([engine.parameters[('bounds', f'axis_{i}_{edge}')]
                   for edge in ['min', 'max']])
                  for i in range(engine.dimensionality)]

        image_axes = list(self.image_axes.values())
        for i, value in enumerate(self.slices):
            if i not in image_axes:
                bounds[i] = (value, value)

        # TODO: calculate positions based on axis order and slices
        grid_positions = slice_grid(tuple(bounds), self.shape, tuple(image_axes), tuple(self.slices))

        # calculate acquisition function
        posterior_mean_value = engine.optimizer.posterior_mean(grid_positions)['f(x)'].reshape(*self.shape)

        # transpose if necessary
        if self.image_axes['x'] > self.image_axes['y']:
            posterior_mean_value = posterior_mean_value.T

        # assign to data object with lock
        with data.w_lock():
            data.states[self.data_key] = posterior_mean_value


@dataclass(eq=False)
class GPCamHyperparameterPlot(MultiPlot):
    compute_with = Location.AdaptiveEngine
    data_key: ClassVar[str] = 'Hyperparameters'
    name = 'Hyperparameters'
    label_key: InitVar[str] = 'Hyperparameter Labels'
    accumulates = True
    stack_plots: ClassVar[bool] = False

    def compute(self, data, engine: 'GPCAMInProcessEngine'):
        if data.states.get(self.data_key, None) is None:
            data.states[self.data_key] = [[] for i in range(len(engine.optimizer.get_hyperparameters()))]
        # assign to data object with lock
        with data.w_lock():
            for i in range(len(engine.optimizer.get_hyperparameters())):
                data.states[self.data_key][i].append(engine.optimizer.get_hyperparameters()[i])
            data.states[self.label_key] = [f"Hyperparameter #{i+1}" for i in range(len(engine.optimizer.get_hyperparameters()))]

@dataclass(eq=False)
class GPCamHyperparameterLogPlot(MultiPlot):
    compute_with = Location.AdaptiveEngine
    data_key: ClassVar[str] = 'hyperparameter training log'
    name = 'Hyperparameters'
    label_key: InitVar[str] = 'Hyperparameter Labels'
    accumulates = True
    stack_plots: ClassVar[bool] = False

    def update(self, widget, data: 'Data', engine: 'GPCAMInProcessEngine'):
        if 'hyperparameter training log' in data.states:
            plot_data = np.array(data.get(['hyperparameter training log']))
            widget.plot(x=plot_data[:, 0], y=plot_data[:, 1])

            if self.pen_key is None:
                self.colorize(widget, data)



class RawGraph(Image):
    # shape = (50, 50)
    data_key = 'Raw data'
    widget_kwargs: dict = field(default_factory=lambda: dict(invert_y=False))
    transform_to_parameter_space: ClassVar[bool] = False

    def compute(self, data, engine: 'GPCamInProcessEngine'):
        import glob
        from PIL import Image as pImage
        import numpy as np

        raw_path = r"C:\data\raw\*.tif"
        raw_images = glob.glob(raw_path)
        raw_image = np.fliplr(np.asarray(pImage.open(np.random.choice(raw_images))).T)
        with data.w_lock():
            data.states[self.data_key] = raw_image



if __name__ == '__main__':
    from pyqtgraph import mkQApp

    app = mkQApp()
    graph = SliceImageGraph(5, [[0, 1]] * 5, data_key='blah')
    graph.make_widget()
    widget = graph.widget
    widget.show()

    app.exec_()



import asyncio
import time

import numpy as np
from ophyd.sim import SynAxis, Device, Cpt, SynSignalRO, SynSignal
from bluesky.plan_stubs import mv, trigger_and_read, create, stage, checkpoint, mov
from PIL import Image
from scipy import ndimage

from tsuchinoko.adaptive.gpCAM_in_process import GPCAMInProcessEngine
from tsuchinoko.adaptive.random_in_process import RandomInProcess
from tsuchinoko.core import Core, ZMQCore
from tsuchinoko.execution.bluesky_in_process import BlueskyInProcessEngine

import numpy as np
import os
import time
import sys
import h5py
import zmq
import zlib
import pickle

from tsuchinoko.execution.threaded_in_process import ThreadedInProcessEngine

def val2ind(val, an_array):
    return np.argmin(abs(an_array-val), axis=0)

class ir_getdata:

    def __init__(self, fileName, ref_path, print_dir=False, dataName=r'./gpdata/Test'):
        h5_groups = []
        epsilon, b = 0.4, 0.29
        with h5py.File(fileName,'r') as f:
            root_name = list(f.keys())[0]
            print(root_name)
            h = f[root_name] #folder name
            self.xy = f[root_name + '/data/xy'][:,:]
            self.wav = f[root_name + '/data/wavenumbers'][:]
            self.N_w = len(self.wav)
            self.data_list = f[root_name + '/data/spectra'][:,:]
            self.data_cube = f[root_name + '/data/image/image_cube'][:, : ,:]
            self.data_list = -np.log10(self.data_list/100 + epsilon) + b
            self.data_cube = -np.log10(self.data_cube/100 + epsilon) + b
            h.visit(h5_groups.append)
            if print_dir:
                print(*h5_groups, sep='\n')
        self.n_spec = 0
        self.cum_spectra = np.zeros((0, self.N_w))
        path = os.path.dirname(dataName)
        self.cum_specName = os.path.join(path, 'all_specs.npy')
        self.ref_spec = np.loadtxt(ref_path, delimiter=',')
        print(f'ref_spec: {os.path.basename(ref_path)}')

        # get X, Y coordinates
        self.X = np.unique(self.xy[:,0])
        self.Y = np.unique(self.xy[:,1])

    def get_spec(self, pos):
        row, col = [], []
        pos = pos.reshape(-1,2)
        for i in range(len(pos)):
            row.append(val2ind(pos[i, 1], self.Y))  # align y coordinate to get row
            col.append(val2ind(pos[i, 0], self.X))  # align x coordinate to get col
        spec = self.data_cube[row, col, :]
        self.cum_spectra = np.append(self.cum_spectra, spec.reshape(1, -1), axis=0)
        np.save(self.cum_specName, self.cum_spectra)
        self.n_spec += 1

    def alignWithRef(self, spec, ref):
        if ref.ndim == 1:
            return spec, ref
        elif ref.ndim == 2:
            pos2, pos1 = val2ind(self.wav, ref[-1, 0]) + 1, val2ind(self.wav, ref[0, 0])
            if spec.ndim == 1: spec = spec.reshape(1, -1)
            if len(self.wav[pos1: pos2]) == ref.shape[0]:
                return spec[:, pos1: pos2], ref[:, 1]
            elif len(self.wav[pos1: pos2]) == ref[::2, :].shape[0]:
                return spec[:, pos1: pos2], ref[::2, 1]
            elif len(self.wav[pos1: pos2][::4]) == ref.shape[0]:
                return spec[:, pos1: pos2: 4], ref[:, 1]
            elif len(self.wav[pos1: pos2][::4]) == ref[::2, :].shape[0]:
                return spec[:, pos1: pos2: 4], ref[::2, 1]
            else:
                print("Length doesn't match")
                return None

    def gp_getCorrData(self, idx):
        spec = self.cum_spectra[idx, :]
        spec, ref = self.alignWithRef(spec, self.ref_spec)
        return np.corrcoef(spec, ref[None, :])[0, 1]

def instrument_synthetic(data):
    # load last saved data
    variance = 1e-3

    if (ir_data.n_spec == 0) and os.path.exists(ir_data.cum_specName) :
        ir_data.cum_spectra = np.load(ir_data.cum_specName)
        ir_data.n_spec = ir_data.cum_spectra.shape[0]

    for idx_data in range(len(data)):
        if data[idx_data]["measured"] == True: continue
        x1 = data[idx_data]["position"][0]
        x2 = data[idx_data]["position"][1]
        pos = np.array([x1, x2])
        ir_data.get_spec(pos)

    for idx_data in range(len(data)):
        if data[idx_data]["measured"] == True: continue
        data[idx_data]["value"] = ir_data.gp_getCorrData(-1)  # Modified for single-point operation idx_data)
        data[idx_data]["variance"] = variance
        data[idx_data]["measured"] = True
        data[idx_data]["time stamp"] = time.time()
    return data

fileName = 'E:\\data\\IR-gpcam\\20180216r_S1_tr_area1.h5'
ref_path = 'E:\\data\\IR-gpcam\\Lipopolysaccharides.CSV'
dataName = 'E:\\data\\IR-gpcam\\'
# ir_data = ir_getdata(fileName, ref_path, dataName=dataName)

## for beamline experiment

def alignWithRef(wav, spec, ref):
    if ref.ndim == 1:
        return spec, ref
    elif ref.ndim == 2:
        pos2, pos1 = val2ind(wav, ref[-1, 0]) + 1, val2ind(wav, ref[0, 0])
        if spec.ndim == 1: spec = spec.reshape(1, -1)
        if len(wav[pos1: pos2]) == ref.shape[0]:
            return spec[:, pos1: pos2], ref[:, 1]
        elif len(wav[pos1: pos2]) == ref[::2, :].shape[0]:
            return spec[:, pos1: pos2], ref[::2, 1]
        elif len(wav[pos1: pos2][::4]) == ref.shape[0]:
            return spec[:, pos1: pos2: 4], ref[:, 1]
        elif len(wav[pos1: pos2][::4]) == ref[::2, :].shape[0]:
            return spec[:, pos1: pos2: 4], ref[::2, 1]
        else:
            print("Length doesn't match")
            return None


def getCorrData(wav, spec, ref):
    spec, ref = alignWithRef(wav, spec, ref)
    return np.corrcoef(spec, ref[None, :])[0, 1]


class zmq_connect:

    def __init__(self, ref_file=r'./ref_spec/black_carbons_on_minerals.CSV', specs_file=r'./gpFiles/spectra.npy',
                 port="6888"):

        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.PAIR)
        self.socket.bind("tcp://*:%s" % port)
        self.ref_spec = np.loadtxt(ref_file, delimiter=',')
        self.cum_spectra = None
        self.specs_file = specs_file
        print(f'Reference spectrum: {os.path.basename(ref_file)}')
        print(f'Raw spectra are saved at: {specs_file}')

    def send_zipped_pickle(self, obj, flags=0, protocol=-1):
        """pack and compress an object with pickle and zlib."""
        pobj = pickle.dumps(obj, protocol)
        zobj = zlib.compress(pobj)
        #         print('zipped pickle is %i bytes' % len(zobj))
        self.socket.send(zobj, flags=flags)

    def recv_zipped_pickle(self, flags=0):
        """reconstruct a Python object sent with zipped_pickle"""
        zobj = self.socket.recv(flags)
        pobj = zlib.decompress(zobj)
        return pickle.loads(pobj)

    def comm_via_zmq_transform(self, data):
        self.send_zipped_pickle(data)
        np.save("data_last_sent", data)
        print("=================================")
        data = self.recv_zipped_pickle()
        for idx in range(len(data)):
            wav, spec = data[idx]["value"]
            if self.cum_spectra is None:
                self.cum_spectra = wav.reshape(1, -1)
            self.cum_spectra = np.append(self.cum_spectra, spec.reshape(1, -1), axis=0)
            data[idx]["value"] = getCorrData(wav, spec, self.ref_spec)
        print("gpCAM received data point#: ", self.cum_spectra.shape[0] - 1)
        np.save("data_last_received", data)
        np.save(self.specs_file, self.cum_spectra)
        return data

    def start(self):
        msg = self.recv_zipped_pickle()
        print(msg)

    def stop(self):
        self.socket.close()
        self.context.term()
        print('zmq connection is stopped.')


if __name__ == '__main__':

    zmq_con = zmq_connect()
    zmq_con.start()

    def measure_target(target, instrument=zmq_con):
        # data = instrument_synthetic([{'position': target, 'measured': False}])
        data = instrument.comm_via_zmq_transform([{'position': target, 'measured': False}])
        return data[0]['value'], data[0]['variance']


    adaptive = GPCAMInProcessEngine(dimensionality=2,
                                    parameter_bounds=[[-20590.4,  -20010.4], [3394.0, 3894.0]],
                                    hyperparameters=[1, 1, 1],
                                    hyperparameter_bounds=[[0.0001,10],[0.001,1e4],[0.001,1e4]])

    execution = ThreadedInProcessEngine(measure_target)

    core = ZMQCore()
    core.set_adaptive_engine(adaptive)
    core.set_execution_engine(execution)

    core.main()



from pathlib import Path

import numpy as np
from PIL import Image
from scipy import ndimage

from tsuchinoko.adaptive.fvgp_gpCAM_in_process import FvgpGPCAMInProcessEngine
from tsuchinoko.core import ZMQCore, CoreState
from tsuchinoko.execution.simple import SimpleEngine

# NOTES:
# 2 signal variances per task

# Load data from a jpg image to be used as a luminosity map
images = [np.flipud(np.asarray(Image.open(Path(__file__).parent / f'peak{i + 1}.png'))) for i in range(2)]
luminosity = [np.average(image, axis=2) for image in images]


# Bilinear sampling will be used to effectively smooth pixel edges in source data
def bilinear_sample(pos):
    print(pos, ndimage.map_coordinates(luminosity, [[0, 1], [pos[1]] * 2, [pos[0]] * 2], order=1), 1, {})
    return pos, np.asarray(ndimage.map_coordinates(luminosity, [[0, 1], [pos[1]] * 2, [pos[0]] * 2], order=1)), (1, 1), {}


execution = SimpleEngine(measure_func=bilinear_sample)

# Define a gpCAM adaptive engine with initial parameters
adaptive = FvgpGPCAMInProcessEngine(dimensionality=2,
                                    output_number=2,
                                    parameter_bounds=[(0, images[0].shape[1]),
                                                      (0, images[0].shape[0])],
                                    hyperparameters=[255**2, 1, 1, 1],
                                    hyperparameter_bounds=[(1e-1, 1e5)]*4)

# Construct a core server
core = ZMQCore()
core.set_adaptive_engine(adaptive)
core.set_execution_engine(execution)

if __name__ == '__main__':
    # Start the core server
    core.state = CoreState.Starting
    core.main()



import sys

import numpy as np

from gpcam.gp_optimizer import  fvGPOptimizer
from .gpCAM_in_process import GPCAMInProcessEngine
from ..graphs.common import GPCamPosteriorCovariance, GPCamAcquisitionFunction, GPCamPosteriorMean, Table, \
    GPCamHyperparameterPlot, Score


class FvgpGPCAMInProcessEngine(GPCAMInProcessEngine):
    """
    A multi-task adaptive engine powered by gpCAM: https://gpcam.readthedocs.io/en/latest/
    """

    def __init__(self, dimensionality, output_number, parameter_bounds, hyperparameters, hyperparameter_bounds, **kwargs):
        self.kwargs = kwargs
        self.output_number = output_number
        super(FvgpGPCAMInProcessEngine, self).__init__(dimensionality, parameter_bounds, hyperparameters, hyperparameter_bounds, **kwargs)

        if dimensionality == 2:
            self.graphs = [GPCamPosteriorCovariance(),
                           GPCamAcquisitionFunction(),
                           GPCamPosteriorMean(),
                           GPCamHyperparameterPlot(),
                           Score(),
                           Table()]
        elif dimensionality > 2:
            self.graphs = [GPCamPosteriorCovariance(),
                           Table()]

    # TODO: refactor this into base
    def init_optimizer(self):
        opts = self.gp_opts.copy()
        if sys.platform == 'darwin':
            opts['compute_device'] = 'numpy'

        hyperparameters = np.asarray([self.parameters[('hyperparameters', f'hyperparameter_{i}')]
                                      for i in range(self.num_hyperparameters)])

        self.optimizer = fvGPOptimizer(init_hyperparameters=hyperparameters,
                                       **opts)

    def _set_hyperparameter(self, parameter, value):
        self.optimizer.gp_initialized = False  # Force re-initialization
        self.optimizer.init_fvgp(np.asarray([self.parameters[('hyperparameters', f'hyperparameter_{i}')]
                                           for i in range(self.num_hyperparameters)]))

    def request_targets(self, position, **kwargs):
        kwargs.update({'x_out': np.arange(self.output_number)})
        return super().request_targets(position, **kwargs)


from pathlib import Path
import numpy as np
from PIL import Image
from scipy import ndimage

from tsuchinoko import examples
from tsuchinoko.adaptive.gpCAM_in_process import GPCAMInProcessEngine
from tsuchinoko.core import ZMQCore
from tsuchinoko.execution.simple import SimpleEngine

# Load data from a jpg image to be used as a luminosity map
image = np.flipud(np.asarray(Image.open(Path(examples.__file__).parent/'sombrero_pug.jpg')))
luminosity = np.average(image, axis=2)


# Bilinear sampling will be used to effectively smooth pixel edges in source data
def bilinear_sample(pos):
    return pos, ndimage.map_coordinates(luminosity, [[pos[1]], [pos[0]]], order=1)[0], 1, {}


execution = SimpleEngine(measure_func=bilinear_sample)


# Define a gpCAM adaptive engine with initial parameters
adaptive = GPCAMInProcessEngine(dimensionality=2,
                                parameter_bounds=[(0, image.shape[1]),
                                                  (0, image.shape[0])],
                                hyperparameters=[255, 100, 100],
                                hyperparameter_bounds=[(.1, 1e5),
                                                       (.1, 1e5),
                                                       (.1, 1e5)])

# Construct a core server
core = ZMQCore()
core.set_adaptive_engine(adaptive)
core.set_execution_engine(execution)

if __name__ == '__main__':
    # Start the core server
    core.main()



import ctypes
import importlib
import os
import runpy
import sys


import click
from pyqtgraph import mkQApp

from . import parameters  # registers parameter types
from . import patches
from .utils import runengine
try:
    from ._version import __version__
except (ImportError, ModuleNotFoundError) as ex:
    raise ImportError("You probably haven't installed tsuchinoko yet: pip install -e .") from ex


@click.command()
@click.argument('core_address', required=False, default='localhost')
def launch_client(core_address='localhost'):
    if os.name == 'nt':
        # https://stackoverflow.com/questions/67599432/setting-the-same-icon-as-application-icon-in-task-bar-for-pyqt5-application
        myappid = 'camera.tsuchinoko'  # arbitrary string
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)  # Allows taskbar icon to be shown on windows

    from .widgets.mainwindow import MainWindow
    qapp = mkQApp('Tsuchinoko')

    main_window = MainWindow(core_address)
    main_window.show()

    sys.exit(qapp.exec_())


@click.command()
@click.argument('demo_name', required=False, default='server_demo')
def launch_server(demo_name='server_demo'):
    demo_module = importlib.import_module(f'tsuchinoko.examples.{demo_name}')
    demo_module.core.main()


@click.command(context_settings=dict(ignore_unknown_options=True))
@click.argument('path', required=True)
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
def bootstrap(path, args):
    """A pyinstaller trick to allow launch of python scripts from built exes"""
    print(path)
    sys.argv.pop(0)
    runpy.run_path(path, {}, "__main__")



import numpy as np
from PySide2.QtCore import Qt
from PySide2.QtWidgets import QApplication, QWidget, QHBoxLayout, QSlider
from ophyd.sim import SynAxis, SynSignal
from scipy.stats import multivariate_normal

from tsuchinoko.main import alignment_plan
from tsuchinoko.utils import runengine

if __name__ == '__main__':
    import pyqtgraph as pg

    min1 = -3
    max1 = 3
    min2 = -3
    max2 = 3
    pinhole1min = -3
    pinhole1max = 3
    pinhole2min = -3
    pinhole2max = 3
    pinhole_min_step = .01

    motor1 = SynAxis(name='motor1', labels={'motors'})
    motor2 = SynAxis(name='motor2', labels={'motors'})
    pinhole1 = SynAxis(name='pinhole1', labels={'motors'})
    pinhole2 = SynAxis(name='pinhole2', labels={'motors'})


    def measure_monitor():
        beam_center = motor1.readback.get(), motor2.readback.get()
        beam_stddev = motor1.readback.get() ** 2 + .5, motor2.readback.get() ** 2 + .5
        pinhole = pinhole1.readback.get(), pinhole2.readback.get()
        # print(beam_center, beam_stddev, pinhole)

        return multivariate_normal.pdf(pinhole,
                                       beam_center,
                                       np.diag(np.asarray(beam_stddev) ** 2))  # * np.random.rand()*1e-1


    monitor = SynSignal(name='monitor',
                        labels={'monitor'},
                        func=measure_monitor)

    qapp = QApplication([])

    w = QWidget()
    w.setLayout(QHBoxLayout())
    iv = pg.PlotWidget()
    scatter = pg.ScatterPlotItem(x=[0], y=[0], size=10, pen=pg.mkPen(None), brush=pg.mkBrush(255, 255, 255, 120))
    arrow = pg.CurveArrow(scatter)
    text = pg.TextItem()
    slider = QSlider(orientation=Qt.Vertical)
    slider.setMinimum(-100)
    slider.setMaximum(100)

    w.layout().addWidget(iv)
    w.layout().addWidget(slider)

    iv.addItem(scatter)
    iv.addItem(arrow)
    iv.addItem(text)
    w.show()
    N = 1000

    RE = runengine.get_run_engine()

    plan = alignment_plan(monitor, motor1, min1, max1, motor2, min2, max2, pinhole1, pinhole1min, pinhole1max, pinhole2,
                          pinhole2min, pinhole2max, pinhole_min_step, 10000)
    RE(plan)

    qapp.exec_()




from time import perf_counter

import numpy as np
from qtpy.QtCore import Qt, QTimer
from loguru import logger
from qtpy.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QComboBox
from pyqtgraph import InfiniteLine, mkPen, PlotWidget, HistogramLUTWidget, mkBrush, functions as fn, FileDialog

from tsuchinoko.graphics_items.clouditem import CloudItem
from tsuchinoko.graphics_items.indicatoritem import BetterCurveArrow
from tsuchinoko.graphics_items.mixins import ClickRequesterPlot

last_export_directory = None


class CloudWidget(QWidget):
    noRepeatKeys = [
        Qt.Key.Key_Right,
        Qt.Key.Key_Left,
        Qt.Key.Key_Up,
        Qt.Key.Key_Down,
        Qt.Key.Key_PageUp,
        Qt.Key.Key_PageDown,
    ]

    def __init__(self, data_key:str, accumulates:bool):
        self.graph = ClickRequesterPlot()
        super().__init__()
        self.data_key = data_key
        self.accumulates = accumulates
        # scatter = ScatterPlotItem(name='scatter', x=[0], y=[0], size=10, pen=mkPen(None), brush=mkBrush(255, 255, 255, 120))
        self.timeline = InfiniteLine(0, pen=mkPen(width=3), movable=True)
        self.timeline.sigPositionChanged.connect(self.timeline_changed)
        self.timeline_plot = PlotWidget()
        self.timeline_plot.hideAxis('left')
        self.timeline_plot.setMouseEnabled(False, False)
        self.timeline_plot.addItem(self.timeline)
        self.timeline_plot.setFixedHeight(40)
        self.cloud = CloudItem(name='scatter', size=10)
        histlut = HistogramLUTWidget()
        histlut.setImageItem(self.cloud)
        self.output_selector = QComboBox()
        self.output_selector.setHidden(True)
        self.output_selector.currentIndexChanged.connect(self.invalidate_cloud)

        self.setLayout(QVBoxLayout())
        self.layout().setContentsMargins(0,0,0,0)
        self.layout().setSpacing(0)
        hlayout = QHBoxLayout()
        self.layout().addLayout(hlayout)
        right_layout = QVBoxLayout()
        right_layout.setSpacing(0)
        right_layout.setContentsMargins(0,0,0,0)


        self.layout().addWidget(self.timeline_plot)
        hlayout.addWidget(self.graph)
        hlayout.addLayout(right_layout)
        right_layout.addWidget(histlut)
        right_layout.addWidget(self.output_selector)

        self.graph.addItem(self.cloud)

        # Hard-coded to show max
        self.max_arrow = BetterCurveArrow(self.cloud.scatter, brush=mkBrush('r'))
        self.last_arrow = BetterCurveArrow(self.cloud.scatter, brush=mkBrush('w'))
        # text = TextItem()

        self.cache = dict()

        self.keysPressed = {}
        self.play_timer = QTimer()
        self.play_rate = 0
        self._paused_play_rate = None
        self.fps = 1  # 1 Hz by default
        self.last_play_time = 0
        self.play_timer.timeout.connect(self.timeout)

    def invalidate_cloud(self):
        self.cloud.clear()

    def timeline_changed(self):
        if not self.cache:
            return
        pos = self.timeline.getXPos()
        self.cloud.updateData(x=self.cache['x'][:int(pos)],
                              y=self.cache['y'][:int(pos)],
                              c=self.cache['v'][:int(pos)],
                              data=self.cache['v'][:int(pos)],
                              hoverable=True,
                              hoverPen=mkPen('b', width=2))

    def nframes(self):
        if self.cache:
            return len(self.cache['v'])

    def keyPressEvent(self, ev):
        if ev.key() == Qt.Key.Key_Space:
            self.toggle_pause()
            ev.accept()
        elif ev.key() == Qt.Key.Key_Home:
            self.set_current_index(0)
            self.play(0)
            ev.accept()
        elif ev.key() == Qt.Key.Key_End:
            self.set_current_index(self.nframes()-1)
            self.play(0)
            ev.accept()
        elif ev.key() in self.noRepeatKeys:
            ev.accept()
            if ev.isAutoRepeat():
                return
            self.keysPressed[ev.key()] = 1
            self.eval_key_state()
        else:
            super().keyPressEvent(ev)

    def keyReleaseEvent(self, ev):
        if ev.key() in [Qt.Key.Key_Space, Qt.Key.Key_Home, Qt.Key.Key_End]:
            ev.accept()
        elif ev.key() in self.noRepeatKeys:
            ev.accept()
            if ev.isAutoRepeat():
                return
            try:
                del self.keysPressed[ev.key()]
            except:
                self.keysPressed = {}
            self.eval_key_state()
        else:
            super().keyReleaseEvent(ev)

    def eval_key_state(self):
        if len(self.keysPressed) == 1:
            key = list(self.keysPressed.keys())[0]
            if key == Qt.Key.Key_Right:
                self.play(20)
                self.jump_frames(1)
                # effectively pause playback for 0.2 s
                self.last_play_time = perf_counter() + 0.2
            elif key == Qt.Key.Key_Left:
                self.play(-20)
                self.jump_frames(-1)
                self.last_play_time = perf_counter() + 0.2
            elif key == Qt.Key.Key_Up:
                self.play(-100)
            elif key == Qt.Key.Key_Down:
                self.play(100)
            elif key == Qt.Key.Key_PageUp:
                self.play(-1000)
            elif key == Qt.Key.Key_PageDown:
                self.play(1000)
        else:
            self.play(0)

    def timeout(self):
        now = perf_counter()
        dt = now - self.last_play_time
        if dt < 0:
            return
        n = int(self.play_rate * dt)
        if n != 0:
            self.last_play_time += (float(n) / self.play_rate)
            if self.timeline.getXPos() + n >= self.nframes():
                self.play(0)
            self.jump_frames(n)

    def play(self, rate=None):
        """Begin automatically stepping frames forward at the given rate (in fps).
        This can also be accessed by pressing the spacebar."""
        if rate is None:
            rate = self._paused_play_rate or self.fps
        if rate == 0 and self.play_rate not in (None, 0):
            self._paused_play_rate = self.play_rate
        self.play_rate = rate

        if rate == 0:
            self.play_timer.stop()
            return

        self.last_play_time = perf_counter()
        if not self.play_timer.isActive():
            self.play_timer.start(abs(int(1000 / rate)))

    def toggle_pause(self):
        if self.play_timer.isActive():
            self.play(0)
        elif self.play_rate == 0:
            if self._paused_play_rate is not None:
                fps = self._paused_play_rate
            else:
                fps = (self.nframes() - 1) / (self.tVals[-1] - self.tVals[0])
            self.play(fps)
        else:
            self.play(self.play_rate)

    def set_current_index(self, ind):
        """Set the currently displayed frame index."""
        index = int(fn.clip_scalar(ind, 0, self.nframes()-1))
        self.timeline.setValue(index)

    def jump_frames(self, n):
        """Move video frame ahead n frames (may be negative)"""
        self.set_current_index(self.timeline.getXPos() + n)

    def update_data(self, data, update_slice: slice):
        # require_clear = False
        timeline_at_end = not self.cache or self.timeline.getXPos() == len(self.cache['v']) - 1

        with data.r_lock():
            v = np.asarray(data[self.data_key].copy())
            x, y = zip(*data.positions)

        if v.ndim == 2:
            if len(v[0]) > 1:
                self.output_selector.setHidden(False)
                if self.output_selector.count() != len(v[0]):
                    self.output_selector.clear()
                    for i in range(len(v[0])):
                        self.output_selector.addItem(f'{i}')
                    self.output_selector.setCurrentIndex(0)
                v = v[:, self.output_selector.currentIndex()]
            else:
                try:
                    v = np.squeeze(v, 1)
                except ValueError:
                    pass

        lengths = len(v), len(x), len(y)
        min_length = min(lengths)

        if not np.all(np.array(lengths) == min_length):
            logger.warning(f'Ragged arrays passed to cloud item with lengths (v, x, y): {lengths}')
            x = x[:min_length]
            y = y[:min_length]
            v = v[:min_length]

        if not len(x):
            return

        self.cache = {'x': x, 'y': y, 'v': v}

        max_index = np.argmax(v)
        last_data_size = min(update_slice.start, len(self.cloud.cData))

        if last_data_size == 0 and timeline_at_end:
            action = self.cloud.setData
        elif not self.accumulates:
            action = self.cloud.updateData
        elif timeline_at_end:
            action = self.cloud.extendData
            x = x[last_data_size + 1:]
            y = y[last_data_size + 1:]
            v = v[last_data_size + 1:]
        else:
            action = None

        if action:
            action(x=x,
                   y=y,
                   c=v,
                   data=v,
                   # size=5,
                   hoverable=True,
                   # hoverSymbol='s',
                   # hoverSize=6,
                   hoverPen=mkPen('b', width=2),
                   # hoverBrush=mkBrush('g'),
                   )
        # scatter.setData(
        #     [{'pos': (xi, yi),
        #       'size': (vi - min(v)) / (max(v) - min(v)) * 20 + 2 if max(v) != min(v) else 20,
        #       'brush': mkBrush(color=mkColor(255, 255, 255)) if i == len(x) - 1 else mkBrush(
        #           color=mkColor(255 - c, c, 0)),
        #       'symbol': '+' if i == len(x) - 1 else 'o'}
        #      for i, (xi, yi, vi, c) in enumerate(zip(x, y, v, c))])

        self.max_arrow.setIndex(max_index)
        self.last_arrow.setIndex(len(self.cloud.cData) - 1)

        self.timeline_plot.setXRange(0, min_length-1, padding=0.01)
        self.timeline.setBounds([0, min_length-1])
        if timeline_at_end:
            with fn.SignalBlock(self.timeline.sigPositionChanged, self.timeline_changed):
                self.timeline.setPos(min_length-1)

        # text.setText(f'Max: {v[max_index]:.2f} ({x[max_index]:.2f}, {y[max_index]:.2f})')
        # text.setPos(x[max_index], y[max_index])

    def getView(self):
        return self.graph


import logging
import time

import numpy as np
from PySide2.QtWidgets import QApplication
from bluesky.plan_stubs import mov, checkpoint, trigger_and_read
from ophyd.sim import SynAxis, SynSignal
from scipy.stats import multivariate_normal

from tsuchinoko.experiment import GPExperiment
from tsuchinoko.widgets.displays import RunEngineControls, Configuration
from tsuchinoko.widgets.mainwindow import MainWindow


class AlignmentExperiment(GPExperiment):
    bounds = np.array([[-.1, .1],
                       [-.1, .1]])
    hyperparameters = np.array([1, 1, 1])
    hyperparameter_bounds = np.array([[.1, 1],  # bounds square of 'value' range
                                      [0.01, 10],
                                      [0.01, 10]])
    max_N = 1e9

    # Devices
    motors = [motor1 := SynAxis(name='motor1', labels={'motors'}), motor2 := SynAxis(name='motor2', labels={'motors'})]

    # inject aliases to mock EpicsMotor
    for device in [motor1, motor2]:
        device.user_readback = device.readback

    def measure_monitor(self):
        beam_center = self.motor1.readback.get(), self.motor2.readback.get()
        beam_stddev = self.motor1.readback.get() ** 2 + .5, self.motor2.readback.get() ** 2 + .5

        return 10 * multivariate_normal.pdf((0, 0),
                                            beam_center,
                                            np.diag(np.asarray(beam_stddev) ** 2))  # * np.random.rand()*1e-1

    def __init__(self, *args, **kwargs):
        super(AlignmentExperiment, self).__init__(*args, **kwargs)

        self.monitor = SynSignal(name='monitor',
                                 labels={'monitor'},
                                 func=self.measure_monitor)

        self.beam_stddev_x = SynSignal(name='beam_stddev_x',
                                       labels={'beam_stddev_x'},
                                       func=lambda: self.motor1.readback.get() ** 2 + .5)
        self.beam_stddev_y = SynSignal(name='beam_stddev_y',
                                       labels={'beam_stddev_y'},
                                       func=lambda: self.motor2.readback.get() ** 2 + .5)

    def acq_func(self, x, gp):
        m = gp.posterior_mean(x)["f(x)"]
        v = gp.posterior_covariance(x)["v(x)"]
        # print('val:', m, v)
        return m + (Configuration().posterior_weight_factor.value()) * np.sqrt(v)

    def instrument_func(self, data, monitor, motor1, motor2):
        for entry in data:
            yield from checkpoint()
            cycle_start = time.time()

            next_motor_position = entry['position']
            yield from mov(motor1, next_motor_position[0], motor2, next_motor_position[1])

            ret = (yield from trigger_and_read([self.monitor, self.beam_stddev_x, self.beam_stddev_y], name='primary'))
            amplitude = ret['monitor']['value']
            stddev_x = ret['beam_stddev_x']['value']
            stddev_y = ret['beam_stddev_y']['value']

            # get value
            metric_factors = [Configuration().mean_weight.value(), Configuration().stddev_x_weight.value(), Configuration().stddev_y_weight.value()]
            metric_vec = np.asarray([amplitude, 1 / stddev_x, 1 / stddev_y]) * np.asarray(metric_factors)
            logging.info(msg=f'metrics: {metric_vec}')
            entry['value'] = np.linalg.norm(metric_vec)
            entry['variance'] = 0.0005 ** 2  # TODO: use the variance from the fit
            entry['metrics'] = dict()
            entry['metrics']['amplitude'] = amplitude
            entry['metrics']['X stddev'] = stddev_x
            entry['metrics']['Y stddev'] = stddev_y
            entry['metrics']['amplitude weighted'] = metric_vec[0]
            entry['metrics']['X stddev weighted (inverted)'] = metric_vec[1]
            entry['metrics']['Y stddev weighted (inverted)'] = metric_vec[2]

            RunEngineControls().measurement_time.setText(f'{time.time() - cycle_start:.1f} s')
        return data

    @property
    def plan(self):
        yield from super(AlignmentExperiment, self).plan(self.motors,
                                                         self.bounds,
                                                         [self.monitor, self.beam_stddev_x, self.beam_stddev_y],
                                                         self.hyperparameters,
                                                         self.hyperparameter_bounds,
                                                         self.max_N,
                                                         monitor=self.monitor,
                                                         motor1=self.motor1,
                                                         motor2=self.motor2,
                                                         )


if __name__ == '__main__':
    import pyqtgraph as pg

    qapp = QApplication([])

    main_window = MainWindow()

    main_window.experiment = AlignmentExperiment(main_window.graph_manager_widget)

    main_window.show()

    iv = pg.PlotWidget()
    scatter = pg.ScatterPlotItem(x=[0], y=[0], size=10, pen=pg.mkPen(None), brush=pg.mkBrush(255, 255, 255, 120))
    arrow = pg.CurveArrow(scatter)
    text = pg.TextItem()
    iv.addItem(scatter)
    iv.addItem(arrow)
    iv.addItem(text)

    exit(qapp.exec_())



from queue import Queue, Empty

from bluesky.plan_stubs import open_run, null

from tsuchinoko.utils.runengine import get_run_engine
from . import Engine


class BlueskyInProcessEngine(Engine):
    """
    An execution engine powered by Bluesky. This engine utilized Bluesky's "plan" concept for experimental procedures,
    and measurements are natively compatible with Databroker ingestion.

    https://nsls-ii.github.io/bluesky/index.html
    """
    def __init__(self, measure_target, get_position):
        # These would normally be on the remote end
        self.targets = Queue()
        self.RE = get_run_engine()
        self.RE(self.target_queue_plan(measure_target, get_position))
        self.position = None
        self.new_measurements = []

    def update_targets(self, positions):
        with self.targets.mutex:
            self.targets.queue.clear()

        for position in positions:
            self.targets.put(position)

    def target_queue_plan(self, measure_target, get_position):
        yield from open_run()
        self.position = tuple((yield from get_position()))
        while True:
            try:
                target = tuple(self.targets.get(timeout=.1))
            except Empty:
                yield from null()
            else:
                self.position = target
                value, variance = (yield from measure_target(target))
                self.new_measurements.append((self.position, value, variance, {}))  # TODO: Add variance; TODO: add metrics

    def get_position(self):
        return self.position

    def get_measurements(self):
        measurements = self.new_measurements
        self.new_measurements = []
        return measurements



import sys


class StdOutPatch:
    def flush(self):
        pass

    def write(self, *args, **kwargs):
        pass


# Jupyter kernels try to flush stdout/stderr, but these might not exist in all environments
for channel in ['stdout', 'stderr']:
    if getattr(sys, channel) is None:
        setattr(sys, channel, StdOutPatch())



import asyncio
import logging
import time
from dataclasses import dataclass, field
from functools import wraps
from queue import PriorityQueue, Empty
from typing import Any
import dask.distributed  # this insulates from errors related to dask asserting its own EventLoopPolicy as squashing the event loop setup for bluesky

from bluesky import RunEngine
from bluesky.utils import DuringTask, RunEngineInterrupted
from qtpy.QtCore import QObject, Signal

from tsuchinoko.utils import threads


@dataclass(order=True)
class PrioritizedPlan:
    priority: int
    args: Any = field(compare=False)


class QRunEngine(QObject):
    sigDocumentYield = Signal(str, dict)
    sigAbort = Signal()  # TODO: wireup me
    sigException = Signal(Exception)
    sigFinish = Signal()
    sigStart = Signal()
    sigPause = Signal()
    sigResume = Signal()
    sigReady = Signal()

    def __init__(self, **kwargs):
        super(QRunEngine, self).__init__()

        self._request_resume = False
        self._kwargs = kwargs
        self._RE = None

        # # TODO: pull from settings plugin
        # from suitcase.mongo_normalized import Serializer
        # #TODO create single databroker db
        # #python-dotenv stores name-value pairs in .env (add to .gitginore)
        # username=os.getenv("USER_MONGO")
        # pw = os.getenv("PASSWD_MONGO")
        # try:
        #     self._RE.subscribe(Serializer(f"mongodb://{username}:{pw}@localhost:27017/mds?authsource=mds",
        #                                  f"mongodb://{username}:{pw}@localhost:27017/fs?authsource=fs"))
        # except OperationFailure as err:
        #     msg.notifyMessage("Could not connect to local mongo database.",
        #                       title="xicam.Acquire Error",
        #                       level=msg.ERROR)
        #     msg.logError(err)

        self.sigFinish.connect(self._check_if_ready)

        self.queue = PriorityQueue()
        self.process_queue_thread = threads.QThreadFutureIterator(self.process_queue,
                                                                  # finished_slot=self._close,
                                                                  interrupt_callable=self._close_RE,
                                                                  name='runengine-qthread')
        self.process_queue_thread.start()

    @property
    def RE(self):
        return self._RE

    def _close_RE(self):
        if self._RE.state != 'idle':
            self._RE.abort('Application is closing.')

    # def _close(self):
    #     self.asyncio_loop.close()

    def process_queue(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        self._RE = RunEngine(context_managers=[], during_task=DuringTask(), loop=self.loop, **self._kwargs)
        self._RE.subscribe(self.sigDocumentYield.emit)

        while True:
            yield
            # get a plan, wait up to .1 sec before hot looping
            if self._RE.state == 'idle':
                try:
                    priority_plan = self.queue.get(block=True, timeout=.1)  # timeout is arbitrary, we'll come right back
                except Empty:
                    continue
                priority, (args, kwargs) = priority_plan.priority, priority_plan.args
            elif self._RE.state == 'paused' and self._request_resume:
                pass
            else:
                time.sleep(.1)
                continue

            self.sigStart.emit()
            try:
                if self._RE.state == 'idle':
                    self._RE(*args, **kwargs)
                elif self._RE.state == 'paused' and self._request_resume:
                    self._request_resume = False
                    self._RE.resume()
            except RunEngineInterrupted:
                logging.critical("Run has been aborted by the user.")
            except RuntimeError as ex:
                logging.error("An error occured during a Bluesky plan. See the Tsuchinoko log for details.")
                logging.error(ex, exc_info=True)
                self.sigException.emit(ex)
            else:
                if priority_plan:
                    priority_plan = None
                    self.queue.task_done()
                    self.sigFinish.emit()
                # msg.showReady()

    @wraps(RunEngine.__call__)
    def __call__(self, *args, **kwargs):
        self.put(*args, **kwargs)

    @property
    def isIdle(self):
        return self._RE.state == 'idle'

    def abort(self, reason=''):
        if self._RE.state != 'idle':
            self._RE.abort(reason=reason)
            self.sigAbort.emit()

    def pause(self, defer=False):
        if self._RE.state != 'paused':
            self._RE.request_pause(defer)
            self.sigPause.emit()

    def resume(self, ):
        if self._RE.state == 'paused':
            self._request_resume = True
            self.sigResume.emit()

    def put(self, *args, priority=1, **kwargs):
        # handle ParameterizedPlan's
        # plan = args[0]
        # if isinstance(args[0], ParameterizedPlan):
        #     # Ask for parameters
        #     param = plan.parameter
        #     if param:
        #         ParameterDialog(param).exec_()

        # reserved = set(kwargs.keys()).union(['plan_type', 'plan_args', 'scan_id', 'time', 'uid'])
        # self._metadata_dialog = MetadataDialog(reserved=reserved)
        # self._metadata_dialog.open()
        # self._metadata_dialog.accepted.connect(partial(self._put, self._metadata_dialog, priority, args, kwargs))
        self._put(priority, args, kwargs)

    def _put(self, priority, args, kwargs):
        self.queue.put(PrioritizedPlan(priority, (args, kwargs)))
        if not self.process_queue_thread.running:
            self.process_queue_thread.start()

    def _check_if_ready(self):
        # RE has finished processing everything in the queue
        if self._RE.state == 'idle' and self.queue.unfinished_tasks == 0:
            self.sigReady.emit()


RE = None


def initialize():
    global RE


def get_run_engine() -> QRunEngine:
    global RE
    if RE is None:
        RE = QRunEngine()
        RE.sigDocumentYield.connect(logging.debug)
    return RE

threads.invoke_as_event(get_run_engine)



"""
Launches a client window; attempts to connect at localhost address
"""
import tsuchinoko
import sys
import faulthandler

if __name__ == '__main__':
    try:
        faulthandler.enable()
    except AttributeError:
        pass

    tsuchinoko.launch_server(sys.argv[1:])



import numpy as np
from caproto.server import PVGroup, pvproperty, template_arg_parser, run
from scipy.stats import multivariate_normal


class TestOpticsBeamline(PVGroup):
    optic_1 = pvproperty(value=0, dtype=float)
    optic_2 = pvproperty(value=0, dtype=float)
    optic_3 = pvproperty(value=3, dtype=float)
    optic_4 = pvproperty(value=4, dtype=float)

    scalar_sensor = pvproperty(value=0, dtype=float, read_only=True)

    target = [1, 2]
    sigma = np.array([1, 1])

    @scalar_sensor.getter
    async def scalar_sensor(self, instance):
        return np.multivariate_normal.pdf([self.optic_1.value, self.optic_2.value],
                                          self.target,
                                          np.diag(self.sigma ** 2)) + 1 * np.random.rand() * 1e-1


class TwoDScanBeamline(TestOpticsBeamline):
    scan_x = pvproperty(value=0, dtype=float)
    scan_y = pvproperty(value=0, dtype=float)

    scan_sensor = pvproperty(value=0, dtype=float, read_only=True)

    @scan_sensor.getter
    async def scan_sensor(self, instance):
        await self.scalar_sensor.read(self.scalar_sensor.data_type)  # update derived value
        return multivariate_normal.pdf([self.scan_x.value, self.scan_y.value],
                                       [0, 0],
                                       np.diag([.1, .1])) * self.scalar_sensor.value


def twod_scan_beamline():
    return TwoDScanBeamline()


if __name__ == "__main__":
    parser, split_args = template_arg_parser(default_prefix='test:', desc='beamline optics test')

    args = parser.parse_args()
    ioc_options, run_options = split_args(args)
    ioc = TwoDScanBeamline(**ioc_options)
    run_options['log_pv_names'] = True
    run(ioc.pvdb, **run_options)



from dataclasses import dataclass
from functools import partial

import numpy as np
from loguru import logger
import numba as nb
from scipy import linalg, sparse
from scipy.ndimage import median_filter

from tsuchinoko.graphics_items.mixins import YInvert, ClickRequester, BetterButtons, LogScaleIntensity, AspectRatioLock, \
    BetterAutoLUTRangeImageView, DomainROI
from tsuchinoko.graphs import Location
from tsuchinoko.graphs.common import Image, ImageViewBlendROI, image_grid, ImageViewBlend, Plot, Bar


class NonViridisBlend(YInvert,
                      ClickRequester,
                      BetterButtons,
                      AspectRatioLock,
                      BetterAutoLUTRangeImageView,
                      # DomainROI,
                      ):
    pass


#@nb.jit
def distance_from_line(x1, x2, pointsx, pointsy):
    """
    Calculate the perpendicular distance from an array of points to a line
    defined by two points, x1 and x2.

    Args:
        points (numpy.ndarray): An Nx2 array of points where each row is a point [p1, p2].
        x1 (numpy.ndarray): A 1x2 array representing the first point on the line [x1_1, x1_2].
        x2 (numpy.ndarray): A 1x2 array representing the second point on the line [x2_1, x2_2].

    Returns:
        numpy.ndarray: An N-element array where each element is the distance of a point to the line.
    """
    points = np.array([pointsx, pointsy]).T
    # Vector from x1 to x2 (direction of the line)
    line_vec = x2 - x1

    # Vectors from x1 to each point
    points_vec = points - x1

    # Cross product of line_vec and points_vec (only z component in 2D)
    cross_prod = np.abs(line_vec[0] * points_vec[:, :, 1] - line_vec[1] * points_vec[:, :, 0])

    # Magnitude of the line vector (length of the line)
    line_mag = np.linalg.norm(line_vec)

    # Distance from points to line (perpendicular distance)
    distances = cross_prod / line_mag

    return distances

def distance_from_line_array(p1, p2, shape):
    return np.fromfunction(partial(distance_from_line,p1,p2), shape)


def projection_operator(x, phi, map_size, center=None, width=1, length=None):
    if not length:
        length = map_size * 2

    if not center:
        center = (map_size/2, map_size/2)

    # get a sampling over a slice through x, phi
    v = np.array([-np.cos(np.deg2rad(phi)), np.sin(np.deg2rad(phi))]) * (-x + map_size/2)
    c = np.array(center)
    f = np.array([-np.sin(np.deg2rad(phi)), -np.cos(np.deg2rad(phi))]) * length / 2

    f0 = v + c + f
    f1 = v + c - f

    x0, y0 = f0
    x1, y1 = f1

    distance = distance_from_line_array(f0-.5, f1-.5, (map_size, map_size))
    projection = np.clip(width-distance, 0, None)

    return projection

@nb.jit()#nb.types.float32[:],(nb.types.float32[:],
                            # nb.types.float32[:, :],
                            # nb.types.optional(nb.types.int64),
                            # nb.types.optional(nb.types.float32[:, :]),
                            # nb.types.optional(nb.types.float32[:])))
def sirt(sinogram, projection_operator, num_iterations=10, inverse_operator=None, initial=None):
    R = np.diag(1 / np.sum(projection_operator, axis=1, dtype=np.float32))
    R = np.nan_to_num(R)
    C = np.diag(1 / np.sum(projection_operator, axis=0, dtype=np.float32))
    C = np.nan_to_num(C)

    if initial is None:
        x_rec = np.zeros(projection_operator.shape[1], dtype=np.float32)
    else:
        x_rec = initial

    for _ in range(num_iterations):
        if inverse_operator is not None:
            x_rec += C @ (inverse_operator @ (R @ (sinogram - projection_operator @ x_rec)))
        else:
            x_rec += C @ (projection_operator.T @ (R @ (sinogram - projection_operator @ x_rec)))

    return x_rec


@dataclass(eq=False)
class ReconstructionGraph(Image):
    compute_with = Location.AdaptiveEngine
    shape: tuple = (32, 32)
    data_key = 'Reconstruction'
    widget_class = NonViridisBlend
    transform_to_parameter_space = False

    def compute(self, data, engine: 'GPCAMInProcessEngine'):
        scores = np.array(data.scores)

        try:
            num_sinograms = len(scores[0])
        except TypeError:
            num_sinograms = 1

        # calculate domain maps
        # self.last_recon = sirt(scores.T.ravel(),
        #                        linalg.block_diag(*[engine.optimizer.A] * num_sinograms),
        #                        num_iterations=1,
        #                        initial=getattr(self, 'last_recon', None))
        try:
            last_recon = getattr(engine.optimizer, 'last_recon', None)
        except Exception:
            last_recon = None

        # assign to data object with lock
        if last_recon is not None:
            with data.w_lock():
                data.states[self.data_key] = np.rot90(last_recon.reshape(*self.shape))


@dataclass(eq=False)
class ProjectionMask(Image):
    compute_with = Location.AdaptiveEngine
    shape: tuple = (32, 32)
    data_key = 'Projection Mask'
    widget_class = NonViridisBlend
    transform_to_parameter_space = False

    def compute(self, data, engine: 'GPCAMInProcessEngine'):
        # assign to data object with lock
        with data.w_lock():
            data.states[self.data_key] = np.rot90(projection_operator(*data.positions[-1], self.shape[0]), 3)



@dataclass(eq=False)
class ProjectionOperatorGraph(Image):
    compute_with = Location.AdaptiveEngine
    shape: tuple = (32, 32)
    data_key = 'Projection Operator'
    # widget_class = NonViridisBlend
    transform_to_parameter_space = False

    def compute(self, data, engine: 'GPCAMInProcessEngine'):
        if not hasattr(self, 'A'):
            self.A = np.empty((0,np.prod(self.shape)))
            self._positions_seen = 0

        length_diff = len(data) - self._positions_seen
        if length_diff:
            for position in data.positions[-length_diff:]:
                A_vec = projection_operator(*position, self.shape[0]).reshape(1, self.shape[0] ** 2)
                self.A = np.vstack([self.A, A_vec])
                self._positions_seen += 1

        # assign to data object with lock
        with data.w_lock():
            data.states[self.data_key] = np.rot90(np.sum(self.A, axis=0).reshape(*self.shape),3)


@dataclass(eq=False)
class SinogramSpaceGPCamAcquisitionFunction(Image):
    compute_with = Location.AdaptiveEngine
    shape: tuple = (50, 50)
    data_key = 'Acquisition Function'
    widget_class = ImageViewBlend
    real_space_bounds: tuple = (32, 32)
    transform_to_parameter_space = False

    def compute(self, data, engine: 'GPCAMInProcessEngine'):
        # if len(engine.optimizer.y_data) % 10:  # only compute every 10th measurement
        #     return
        from tsuchinoko.adaptive.gpCAM_in_process import gpcam_acquisition_functions  # avoid circular import

        grid_positions = image_grid(self.real_space_bounds, self.shape)

        # check if acquisition function is accessible
        if engine.parameters['acquisition_function'] not in gpcam_acquisition_functions:
            logger.exception(ValueError('The selected acquisition_function is not available for display.'))
            return

        # calculate acquisition function
        grid_positions = [grid_positions[:len(grid_positions)//2], grid_positions[len(grid_positions)//2:]]
        acquisition_function_value = np.hstack([engine.optimizer.evaluate_acquisition_function(p,
                                                                                    acquisition_function=
                                                                                    gpcam_acquisition_functions[
                                                                                        engine.parameters[
                                                                                            'acquisition_function']],
                                                                                    origin=engine.last_position) for p in grid_positions])

        try:
            acquisition_function_value = acquisition_function_value.reshape(*self.shape)
        except (ValueError, AttributeError):
            acquisition_function_value = np.array([[0]])

        # assign to data object with lock
        with data.w_lock():
            data.states[self.data_key] = acquisition_function_value


@dataclass(eq=False)
class ReconHistogram(Bar):
    compute_with = Location.AdaptiveEngine
    data_key = 'Recon Histogram'
    name = "Recon Histogram"

    def compute(self, data, engine: 'GPCAMInProcessEngine'):
        scores = np.array(data.scores)

        try:
            num_sinograms = len(scores[0])
        except TypeError:
            num_sinograms = 1

        # calculate domain maps
        # self.last_recon = sirt(scores.T.ravel(),
        #                        linalg.block_diag(*[engine.optimizer.A] * num_sinograms),
        #                        num_iterations=1,
        #                        initial=getattr(self, 'last_recon', None))
        try:
            last_recon = getattr(engine.optimizer, 'last_recon', None)
        except Exception:
            last_recon = None

        if last_recon is not None:
            # calculate histogram
            y, x = np.histogram(last_recon, bins=100)

            # assign to data object with lock
            with data.w_lock():
                data.states[self.data_key] = [y, x]


@dataclass(eq=False)
class RealSpacePosteriorMean(Image):
    compute_with = Location.AdaptiveEngine
    shape:tuple = (50, 50)
    data_key = 'Posterior Mean'
    widget_class = ImageViewBlend
    transform_to_parameter_space = False

    def compute(self, data, engine: 'GPCAMInProcessEngine'):
        bounds = tuple(tuple(engine.parameters[('bounds', f'axis_{i}_{edge}')]
                   for edge in ['min', 'max'])
                  for i in range(engine.dimensionality))

        grid_positions = image_grid((bounds[0], bounds[0]), self.shape)
        shape = self.shape

        # if multi-task, extend the grid_positions to include the task dimension
        if hasattr(engine, 'output_number'):
            grid_positions = np.vstack([np.hstack([grid_positions, np.full((grid_positions.shape[0], 1), i)]) for i in range(engine.output_number)])
            shape = (*self.shape, engine.output_number)

        # calculate acquisition function
        posterior_mean_value = np.rot90(np.fliplr(engine.optimizer.posterior_mean(grid_positions)['f(x)'].reshape(*shape)),3)

        # assign to data object with lock
        with data.w_lock():
            data.states['Posterior Mean'] = posterior_mean_value


@dataclass(eq=False)
class RealSpacePosteriorVariance(Image):
    compute_with = Location.AdaptiveEngine
    shape:tuple = (50, 50)
    data_key = 'Posterior Variance'
    widget_class = ImageViewBlend
    transform_to_parameter_space = False

    def compute(self, data, engine: 'GPCAMInProcessEngine'):
        bounds = tuple(tuple(engine.parameters[('bounds', f'axis_{i}_{edge}')]
                   for edge in ['min', 'max'])
                  for i in range(engine.dimensionality))

        grid_positions = image_grid((bounds[0], bounds[0]), self.shape)
        shape = self.shape

        # if multi-task, extend the grid_positions to include the task dimension
        if hasattr(engine, 'output_number'):
            grid_positions = np.vstack([np.hstack([grid_positions, np.full((grid_positions.shape[0], 1), i)]) for i in range(engine.output_number)])
            shape = (*self.shape, engine.output_number)

        # calculate acquisition function
        posterior_variance_value = np.rot90(np.fliplr(engine.optimizer.posterior_covariance(grid_positions)['v(x)'].reshape(*shape)),3)

        # assign to data object with lock
        with data.w_lock():
            data.states['Posterior Variance'] = posterior_variance_value


@dataclass(eq=False)
class InvertedRealSpacePosteriorMean(Image):
    compute_with = Location.AdaptiveEngine
    shape:tuple = (64, 180)
    data_key = 'Posterior Mean'
    widget_class = ImageViewBlend
    transform_to_parameter_space = False
    A_inv:np.array = None

    def compute(self, data, engine: 'GPCAMInProcessEngine'):
        bounds = tuple(tuple(engine.parameters[('bounds', f'axis_{i}_{edge}')]
                   for edge in ['min', 'max'])
                  for i in range(engine.dimensionality))

        grid_positions = image_grid(((0,self.shape[0]), (0, 180)), self.shape)
        shape = self.shape

        # if multi-task, extend the grid_positions to include the task dimension
        if hasattr(engine, 'output_number'):
            grid_positions = np.vstack([np.hstack([grid_positions, np.full((grid_positions.shape[0], 1), i)]) for i in range(engine.output_number)])
            shape = (*self.shape, engine.output_number)

        # calculate posterior_mean
        posterior_mean = engine.optimizer.posterior_mean(grid_positions)['f(x)']

        # invert posterior_mean
        real_space_posterior_mean = np.rot90((self.A_inv @ posterior_mean).reshape(shape[0], shape[0]), 3)

        # assign to data object with lock
        with data.w_lock():
            data.states[self.data_key] = real_space_posterior_mean


@dataclass(eq=False)
class InvertedRealSpacePosteriorVariance(Image):
    compute_with = Location.AdaptiveEngine
    shape:tuple = (64, 180)
    data_key = 'Posterior Variance'
    widget_class = ImageViewBlend
    transform_to_parameter_space = False
    A_inv:np.array = None

    def compute(self, data, engine: 'GPCAMInProcessEngine'):
        bounds = tuple(tuple(engine.parameters[('bounds', f'axis_{i}_{edge}')]
                   for edge in ['min', 'max'])
                  for i in range(engine.dimensionality))

        grid_positions = image_grid(((0,self.shape[0]), (0, 180)), self.shape)
        shape = self.shape

        # if multi-task, extend the grid_positions to include the task dimension
        if hasattr(engine, 'output_number'):
            grid_positions = np.vstack([np.hstack([grid_positions, np.full((grid_positions.shape[0], 1), i)]) for i in range(engine.output_number)])
            shape = (*self.shape, engine.output_number)

        # calculate posterior_ variance
        posterior_variance = engine.optimizer.posterior_covariance(grid_positions)['v(x)']

        # invert posterior_variance
        real_space_posterior_variance = np.rot90((self.A_inv @ posterior_variance).reshape(shape[0], shape[0]), 3)

        # assign to data object with lock
        with data.w_lock():
            data.states[self.data_key] = real_space_posterior_variance

@dataclass(eq=False)
class InvertedSinoSpacePosteriorMean(Image):
    compute_with = Location.AdaptiveEngine
    shape:tuple = (64, 180)
    data_key = 'Sino Posterior Mean'
    widget_class = ImageViewBlend
    transform_to_parameter_space = False
    upsampling: int = 3

    def compute(self, data, engine: 'GPCAMInProcessEngine'):
        grid_shape = self.shape[0] * self.upsampling, self.shape[1] + 1
        grid_positions = image_grid(((0,self.shape[0]), (0, self.shape[1])), grid_shape)

        # # if multi-task, extend the grid_positions to include the task dimension
        # if hasattr(engine, 'output_number'):
        #     grid_positions = np.vstack([np.hstack([grid_positions, np.full((grid_positions.shape[0], 1), i)]) for i in range(engine.output_number)])
        #     shape = (*self.shape, engine.output_number)

        # calculate posterior_mean
        posterior_mean = engine.optimizer.posterior_mean(grid_positions)['f(x)'].reshape(*grid_shape)

        # invert posterior_mean
        real_space_posterior_mean = posterior_mean

        # assign to data object with lock
        with data.w_lock():
            data.states[self.data_key] = real_space_posterior_mean


@dataclass(eq=False)
class InvertedSinoSpacePosteriorVariance(Image):
    compute_with = Location.AdaptiveEngine
    shape:tuple = (64, 180)
    data_key = 'Sino Posterior Variance'
    widget_class = ImageViewBlend
    transform_to_parameter_space = False
    upsampling: int = 3

    def compute(self, data, engine: 'GPCAMInProcessEngine'):
        grid_shape = self.shape[0] * self.upsampling, self.shape[1] + 1
        grid_positions = image_grid(((0,self.shape[0]), (0, self.shape[1])), grid_shape)

        # # if multi-task, extend the grid_positions to include the task dimension
        # if hasattr(engine, 'output_number'):
        #     grid_positions = np.vstack([np.hstack([grid_positions, np.full((grid_positions.shape[0], 1), i)]) for i in range(engine.output_number)])
        #     shape = (*self.shape, engine.output_number)

        # calculate posterior_mean
        posterior_variance = engine.optimizer.posterior_covariance(grid_positions)['v(x)'].reshape(*grid_shape)

        # invert posterior_mean
        real_space_posterior_mean = posterior_variance.reshape(self.shape)

        # assign to data object with lock
        with data.w_lock():
            data.states[self.data_key] = real_space_posterior_mean

import tomopy


@dataclass(eq=False)
class InvertedRecon(Image):
    compute_with = Location.AdaptiveEngine
    shape:tuple = (64, 180)
    data_key = 'Reconstruction'
    widget_class = ImageViewBlend
    transform_to_parameter_space = False
    upsampling:int = 3
    algorithm:str = 'art'
    start_finding_center_at: int = 500
    find_center_every: int = 100

    def compute(self, data, engine: 'GPCAMInProcessEngine'):
        grid_shape = self.shape[0] * self.upsampling, self.shape[1] + 1
        grid_positions = image_grid(((0,self.shape[0]), (0, self.shape[1])), grid_shape)
        # shape = self.shape
        #
        # # if multi-task, extend the grid_positions to include the task dimension
        # if hasattr(engine, 'output_number'):
        #     grid_positions = np.vstack([np.hstack([grid_positions, np.full((grid_positions.shape[0], 1), i)]) for i in range(engine.output_number)])
        #     shape = (*self.shape, engine.output_number)

        # calculate posterior_mean
        sinogram = engine.optimizer.posterior_mean(grid_positions)['f(x)'].reshape(*grid_shape)

        # filter and clip
        sinogram = median_filter(sinogram, 3)
        sinogram = np.clip(sinogram, 3000, None)

        # reconstruct
        theta = np.deg2rad(np.linspace(0, self.shape[1], self.shape[1] + 1))
        last_center_found = data.states.get('recon_last_center_found', 0)
        if 'recon_center' in data.states:
            center = data.states.get('recon_center')
        elif len(data) > self.start_finding_center_at and len(data) - last_center_found > self.find_center_every:
            center = tomopy.find_center(sinogram.T[np.newaxis, :, :],
                                        theta,
                                        init=self.shape[0] * self.upsampling / 2,
                                        ind=0,
                                        tol=0.01,
                                        sinogram_order=True)
            data.states['recon_last_center_found'] = len(data)
        else:
            center = self.shape[0] / 2
        recon = tomopy.recon(sinogram.T[np.newaxis, :, :],
                             theta,
                             center=center * self.upsampling,
                             algorithm=self.algorithm,
                             sinogram_order=True)

        # mask
        recon = tomopy.circ_mask(recon, axis=0, ratio=0.95)

        # assign to data object with lock
        with data.w_lock():
            data.states[self.data_key] = np.rot90(recon.reshape(self.shape[0]*self.upsampling,
                                                                self.shape[0]*self.upsampling), 3)


import time
from typing import List

import numpy as np

from tsuchinoko.adaptive import Data
from tsuchinoko.adaptive.gpCAM_in_process import GPCAMInProcessEngine, acquisition_functions
from tsuchinoko.utils.logging import log_time


class Point:
    """A point located at (x,y) in 2D space.

    Each Point object may be associated with a payload object.

    """

    def __init__(self, x, y, value=None, variance=None, metrics=None, acq_func_value=None):
        self.x, self.y = x, y
        self.value = value
        self.variance = variance
        self.metrics = metrics
        self.acq_func_value = acq_func_value

    def __repr__(self):
        return '{}: {}'.format(str((self.x, self.y)), repr([self.value, self.variance, self.metrics, self.acq_func_value]))
    def __str__(self):
        return 'P({:.2f}, {:.2f})'.format(self.x, self.y)

    def distance_to(self, other):
        try:
            other_x, other_y = other.x, other.y
        except AttributeError:
            other_x, other_y = other
        return np.hypot(self.x - other_x, self.y - other_y)

class Rect:
    """A rectangle centred at (cx, cy) with width w and height h."""

    def __init__(self, cx, cy, w, h):
        self.cx, self.cy = cx, cy
        self.w, self.h = w, h
        self.west_edge, self.east_edge = cx - w/2, cx + w/2
        self.north_edge, self.south_edge = cy - h/2, cy + h/2

    def __repr__(self):
        return str((self.west_edge, self.east_edge, self.north_edge,
                self.south_edge))

    def __str__(self):
        return '({:.2f}, {:.2f}, {:.2f}, {:.2f})'.format(self.west_edge,
                    self.north_edge, self.east_edge, self.south_edge)

    def contains(self, point):
        """Is point (a Point object or (x,y) tuple) inside this Rect?"""

        try:
            point_x, point_y = point.x, point.y
        except AttributeError:
            point_x, point_y = point

        return (point_x >= self.west_edge and
                point_x <  self.east_edge and
                point_y >= self.north_edge and
                point_y < self.south_edge)

    def intersects(self, other):
        """Does Rect object other interesect this Rect?"""
        return not (other.west_edge > self.east_edge or
                    other.east_edge < self.west_edge or
                    other.north_edge > self.south_edge or
                    other.south_edge < self.north_edge)

    def draw(self, ax, c='k', lw=1, **kwargs):
        x1, y1 = self.west_edge, self.north_edge
        x2, y2 = self.east_edge, self.south_edge
        ax.plot([x1,x2,x2,x1,x1],[y1,y1,y2,y2,y1], c=c, lw=lw, **kwargs)

    @property  # TODO: memoize
    def area(self):
        return self.w * self.h

    @property
    def center(self):
        return self.cx, self.cy


class QuadTree:
    """A class implementing a quadtree."""

    def __init__(self, boundary, max_points=4, depth=0):
        """Initialize this node of the quadtree.

        boundary is a Rect object defining the region from which points are
        placed into this node; max_points is the maximum number of points the
        node can hold before it must divide (branch into four more nodes);
        depth keeps track of how deep into the quadtree this node lies.

        """

        self.boundary = boundary
        self.max_points = max_points
        self.points = []
        self.depth = depth
        # A flag to indicate whether this node has divided (branched) or not.
        self.divided = False

    def __str__(self):
        """Return a string representation of this node, suitably formatted."""
        sp = ' ' * self.depth * 2
        s = str(self.boundary) + '\n'
        s += sp + ', '.join(str(point) for point in self.points)
        if not self.divided:
            return s
        return s + '\n' + '\n'.join([
                sp + 'nw: ' + str(self.nw), sp + 'ne: ' + str(self.ne),
                sp + 'se: ' + str(self.se), sp + 'sw: ' + str(self.sw)])

    def smallest_containing_quad(self, point:Point) -> 'QuadTree':
        if self.divided:
            for div in [self.nw, self.ne, self.se, self.sw]:
                if div.boundary.contains(point):
                    return div.smallest_containing_quad(point)
        elif self.boundary.contains(point):
            return self
        else:
            return None

    def divide(self):
        """Divide (branch) this node by spawning four children nodes."""

        cx, cy = self.boundary.cx, self.boundary.cy
        w, h = self.boundary.w / 2, self.boundary.h / 2
        # The boundaries of the four children nodes are "northwest",
        # "northeast", "southeast" and "southwest" quadrants within the
        # boundary of the current node.
        self.nw = QuadTree(Rect(cx - w/2, cy - h/2, w, h),
                                    self.max_points, self.depth + 1)
        self.ne = QuadTree(Rect(cx + w/2, cy - h/2, w, h),
                                    self.max_points, self.depth + 1)
        self.se = QuadTree(Rect(cx + w/2, cy + h/2, w, h),
                                    self.max_points, self.depth + 1)
        self.sw = QuadTree(Rect(cx - w/2, cy + h/2, w, h),
                                    self.max_points, self.depth + 1)
        self.divided = True
        return self.divisions

    def insert(self, point):
        """Try to insert Point point into this QuadTree."""

        if not self.boundary.contains(point):
            # The point does not lie inside boundary: bail.
            return False
        if len(self.points) < self.max_points:
            # There's room for our point without dividing the QuadTree.
            self.points.append(point)
            return True

        # No room: divide if necessary, then try the sub-quads.
        if not self.divided:
            self.divide()

        return (self.ne.insert(point) or
                self.nw.insert(point) or
                self.se.insert(point) or
                self.sw.insert(point))

    def query(self, boundary, found_points):
        """Find the points in the quadtree that lie within boundary."""

        if not self.boundary.intersects(boundary):
            # If the domain of this node does not intersect the search
            # region, we don't need to look in it for points.
            return False

        # Search this node's points to see if they lie within boundary ...
        for point in self.points:
            if boundary.contains(point):
                found_points.append(point)
        # ... and if this node has children, search them too.
        if self.divided:
            self.nw.query(boundary, found_points)
            self.ne.query(boundary, found_points)
            self.se.query(boundary, found_points)
            self.sw.query(boundary, found_points)
        return found_points


    def query_circle(self, boundary, centre, radius, found_points):
        """Find the points in the quadtree that lie within radius of centre.

        boundary is a Rect object (a square) that bounds the search circle.
        There is no need to call this method directly: use query_radius.

        """

        if not self.boundary.intersects(boundary):
            # If the domain of this node does not intersect the search
            # region, we don't need to look in it for points.
            return False

        # Search this node's points to see if they lie within boundary
        # and also lie within a circle of given radius around the centre point.
        for point in self.points:
            if (boundary.contains(point) and
                    point.distance_to(centre) <= radius):
                found_points.append(point)

        # Recurse the search into this node's children.
        if self.divided:
            self.nw.query_circle(boundary, centre, radius, found_points)
            self.ne.query_circle(boundary, centre, radius, found_points)
            self.se.query_circle(boundary, centre, radius, found_points)
            self.sw.query_circle(boundary, centre, radius, found_points)
        return found_points

    def query_radius(self, centre, radius, found_points):
        """Find the points in the quadtree that lie within radius of centre."""

        # First find the square that bounds the search circle as a Rect object.
        boundary = Rect(*centre, 2*radius, 2*radius)
        return self.query_circle(boundary, centre, radius, found_points)


    def __len__(self):
        """Return the number of points in the quadtree."""

        npoints = len(self.points)
        if self.divided:
            npoints += len(self.nw)+len(self.ne)+len(self.se)+len(self.sw)
        return npoints

    def draw(self, ax):
        """Draw a representation of the quadtree on Matplotlib Axes ax."""

        self.boundary.draw(ax)
        if self.divided:
            self.nw.draw(ax)
            self.ne.draw(ax)
            self.se.draw(ax)
            self.sw.draw(ax)

    @property
    def divisions(self):
        if self.divided:
            return self.nw, self.ne, self.se, self.sw
        else:
            return []

    @property
    def children_points(self):
        yield from self.points
        for div in self.divisions:
            yield from div.children_points

class QuadTreeEngine(GPCAMInProcessEngine):

    def __init__(self, parameter_bounds, hyperparameters, hyperparameter_bounds, **kwargs):
        super(QuadTreeEngine, self).__init__(2, parameter_bounds, hyperparameters, hyperparameter_bounds, **kwargs)

    def reset(self):
        parameter_bounds = np.asarray([[self.parameters[('bounds', f'axis_{i}_{edge}')]
                                        for edge in ['min', 'max']]
                                       for i in range(self.dimensionality)])

        cx = (parameter_bounds[0][1] + parameter_bounds[0][0]) / 2
        cy = (parameter_bounds[1][1] + parameter_bounds[1][0]) / 2

        width = parameter_bounds[0][1] - parameter_bounds[0][0]
        height = parameter_bounds[1][1] - parameter_bounds[1][0]

        domain = Rect(cx, cy, width, height)
        self.quadtree = QuadTree(domain)
        self.target_queue: List[QuadTree] = [self.quadtree]
        super(QuadTreeEngine, self).reset()
        self._invalidate_all = False

        self._update_counter = 0

    def request_targets(self, position):
        if not self.target_queue:
            points = list(self.quadtree.children_points)
            acq_func_values = [point.acq_func_value for point in points]  # TODO: avoid list-comp
            # n_largest = np.argpartition(acq_func_values, -n)[-n:]
            largest_point = points[np.argmin(acq_func_values)]  # TODO: do nones affect this?
            target_quad = self.quadtree.smallest_containing_quad(largest_point)
            target_divisions = target_quad.divide()
            target_divisions = [div for div in target_divisions if not div.boundary.contains(largest_point)]

            self.target_queue = target_divisions

        return [(np.random.uniform(div.boundary.west_edge, div.boundary.east_edge),
                 np.random.uniform(div.boundary.south_edge, div.boundary.north_edge)) for div in self.target_queue]

    def update_measurements(self, data: Data):
        with data.r_lock():  # quickly grab values within lock before passing to optimizer
            positions = data.positions.copy()
            scores = data.scores.copy()
            variances = data.variances.copy()

        self.optimizer.tell(positions, scores, variances)

        for target_division in self.target_queue:
            for i, position in enumerate(positions[-4:]):  # only check last 3 points (dim**2 - 1)
                if target_division.boundary.contains(Point(*position)):
                    self.quadtree.insert(Point(*position, scores[-i-1], variances[-i-1]))
                    self.target_queue.remove(target_division)
                    break

        # TODO: this could be limited in range to only points nearby updated points



        positions = [(point.x, point.y) for point in self.quadtree.children_points]

        # calculate acquisition function
        with log_time('updating acq_func values', cumulative_key='updating acq_func values'):

            acquisition_function_values = self.optimizer.evaluate_acquisition_function(positions,
                                                                                      acquisition_function=
                                                                                      acquisition_functions[
                                                                                          self.parameters[
                                                                                              'acquisition_function']])

        for i, point in enumerate(self.quadtree.children_points):
            point.acq_func_value = acquisition_function_values[i]*self.quadtree.smallest_containing_quad(point).boundary.area

        if not len(data)%100:
            import matplotlib.pyplot as plt

            DPI = 72
            # np.random.seed(60)

            # width, height = image.shape[1], image.shape[0]

            N = 15000
            # coords = np.random.randn(N, 2) * height / 3 + (width / 2, height / 2)
            # values = [bilinear_sample(pos) for pos in coords]
            # points = [Point(*coord, payload=value) for coord, value in zip(coords, values)]

            # domain = Rect(width / 2, height / 2, width, height)
            # qtree = QuadTree(domain, 3)
            # for point in points:
            #     qtree.insert(point)

            # print('Number of points in the domain =', len(qtree))

            fig = plt.figure(figsize=(700 / DPI, 500 / DPI), dpi=DPI)
            ax = plt.subplot()
            # ax.set_xlim(0, width)
            # ax.set_ylim(0, height)
            self.quadtree.draw(ax)

            points = list(self.quadtree.children_points)
            ax.scatter([p.x for p in points], [p.y for p in points], s=400, c=[p.value for p in points], cmap='gray')
            # ax.set_xticks([])
            # ax.set_yticks([])

            # centre, radius = (width / 2, height / 2), 120
            # found_points = []
            # qtree.query_radius(centre, radius, found_points)
            # print('Number of found points =', len(found_points))

            # ax.scatter([p.x for p in found_points], [p.y for p in found_points],
            #            facecolors='none', edgecolors='r', s=32)

            # circle = plt.Circle(centre, radius, ec='r')
            # Rect(*centre, 2 * radius, 2 * radius).draw(ax, c='r')

            # ax.invert_yaxis()
            plt.tight_layout()
            # plt.savefig('search-quadtree-circle.png')#, DPI=72)
            plt.show()

    def update_metrics(self, data: Data):
        if self._update_counter>10:
            super().update_metrics(data)
            self._update_counter = 0
        else:
            self._update_counter += 1

    def train(self):
        self._invalidate_all = super().train()



if __name__ == '__main__':
    from scipy import ndimage
    from PIL import Image
    from pathlib import Path

    # Load data from a jpg image to be used as a luminosity map
    image = np.flipud(np.asarray(Image.open(Path(__file__).parent.parent.parent / 'examples' / 'sombrero_pug.jpg')))
    luminosity = np.average(image, axis=2)


    # Bilinear sampling will be used to effectively smooth pixel edges in source data
    def bilinear_sample(pos):

        return pos, ndimage.map_coordinates(luminosity, [[pos[1]], [pos[0]]], order=1)[0], 1, {}


    # Poisson noise applied to measured value; variance is set in accordance with poisson statistics
    def noisy_sample(pos):
        pos, value, variance, extra = bilinear_sample(pos)
        return pos, np.random.poisson(value), value, extra


    def gaussian_sample(pos):
        pos, value, variance, extra = bilinear_sample(pos)
        return pos, np.random.standard_normal(1)[0] * 30 + value, value, extra



    import numpy as np
    import matplotlib.pyplot as plt
    from quadtree import Point, Rect, QuadTree
    from matplotlib import gridspec

    DPI = 72
    # np.random.seed(60)

    width, height = image.shape[1], image.shape[0]

    N = 15000
    coords = np.random.randn(N, 2) * height / 3 + (width / 2, height / 2)
    values = [bilinear_sample(pos) for pos in coords]
    points = [Point(*coord, payload=value) for coord, value in zip(coords, values)]

    domain = Rect(width / 2, height / 2, width, height)
    qtree = QuadTree(domain, 3)
    for point in points:
        qtree.insert(point)

    print('Number of points in the domain =', len(qtree))

    fig = plt.figure(figsize=(700 / DPI, 500 / DPI), dpi=DPI)
    ax = plt.subplot()
    ax.set_xlim(0, width)
    ax.set_ylim(0, height)
    qtree.draw(ax)

    ax.scatter([p.x for p in points], [p.y for p in points], s=400, c=[p.payload[1] for p in points], cmap='gray')
    # ax.set_xticks([])
    # ax.set_yticks([])

    centre, radius = (width / 2, height / 2), 120
    found_points = []
    qtree.query_radius(centre, radius, found_points)
    print('Number of found points =', len(found_points))

    ax.scatter([p.x for p in found_points], [p.y for p in found_points],
               facecolors='none', edgecolors='r', s=32)

    circle = plt.Circle(centre, radius, ec='r')
    Rect(*centre, 2 * radius, 2 * radius).draw(ax, c='r')

    # ax.invert_yaxis()
    plt.tight_layout()
    # plt.savefig('search-quadtree-circle.png')#, DPI=72)
    plt.show()



import sys
import uuid
from functools import cached_property
from typing import Callable

import numpy as np
from pyqtgraph.parametertree.parameterTypes import SimpleParameter, GroupParameter, ListParameter
from loguru import logger

from gpcam.gp_optimizer import GPOptimizer
from . import Engine, Data
from .acquisition_functions import explore_target_100, radical_gradient
from ..graphs.common import Variance, GPCamPosteriorCovariance, Score, GPCamAcquisitionFunction, GPCamPosteriorMean, \
    Table, HighDimensionalityGPCamPosteriorMean, GPCamHyperparameterPlot
from ..parameters import TrainingParameter

gpcam_acquisition_functions = {s: s for s in ['variance', 'shannon_ig', 'ucb', 'maximum', 'minimum', 'covariance', 'gradient', 'explore_target_100']}
gpcam_acquisition_functions['explore_target_100'] = explore_target_100
gpcam_acquisition_functions['radical_gradient'] = radical_gradient


def prepend_update_acquisition_functions(acquisition_functions:dict):
    cpy = gpcam_acquisition_functions.copy()
    gpcam_acquisition_functions.clear()
    gpcam_acquisition_functions.update(acquisition_functions)
    gpcam_acquisition_functions.update(cpy)


class GPCAMInProcessEngine(Engine):
    """
    An adaptive engine powered by gpCAM: https://gpcam.readthedocs.io/en/latest/
    """
    default_retrain_globally_at = (20, 50, 100, 400, 1000)
    default_retrain_locally_at = (20, 40, 60, 80, 100, 200, 400, 1000)
    default_retrain_mcmc_at = tuple()

    def __init__(self, dimensionality, parameter_bounds, hyperparameters, hyperparameter_bounds,
                 acquisition_functions:dict[str, Callable]=None,
                 gp_opts: dict = None, ask_opts: dict = None):
        self.dimensionality = dimensionality
        self.gp_opts = gp_opts or {}
        self.ask_opts = ask_opts or {}
        self.num_hyperparameters = len(hyperparameters)
        if acquisition_functions:
            prepend_update_acquisition_functions(acquisition_functions)

        for i in range(dimensionality):
            for j, edge in enumerate(['min', 'max']):
                self.parameters[('bounds', f'axis_{i}_{edge}')] = parameter_bounds[i][j]
        for i in range(self.num_hyperparameters):
            for j, edge in enumerate(['min', 'max']):
                self.parameters[('hyperparameters', f'hyperparameter_{i}_{edge}')] = hyperparameter_bounds[i][j]
            self.parameters.child('hyperparameters', f'hyperparameter_{i}').setValue(hyperparameters[i], blockSignal=self._set_hyperparameter)

        self.reset()

        if dimensionality == 2:
            self.graphs = [GPCamPosteriorCovariance(),
                           GPCamAcquisitionFunction(),
                           GPCamPosteriorMean(),
                           # GPCamAverageCovariance(),
                           GPCamHyperparameterPlot(),
                           Table(),
                           Variance(),
                           Score()]
        elif dimensionality > 2:
            self.graphs = [GPCamPosteriorCovariance(),
                           HighDimensionalityGPCamPosteriorMean(dimensions=dimensionality, bounds=parameter_bounds),
                           # GPCamAverageCovariance(),
                           GPCamHyperparameterPlot(),
                           Table()]

    def init_optimizer(self):
        opts = self.gp_opts.copy()
        if sys.platform == 'darwin':
            opts['compute_device'] = 'numpy'

        hyperparameters = np.asarray([self.parameters[('hyperparameters', f'hyperparameter_{i}')]
                                      for i in range(self.num_hyperparameters)])

        self.optimizer = GPOptimizer(init_hyperparameters=hyperparameters,
                                     **opts)

    def reset(self):
        self._completed_training = {'global': set(),
                                    'local': set()}
        self.init_optimizer()

    @cached_property
    def parameters(self):
        hyper_parameters = [SimpleParameter(title=f'Hyperparameter #{i + 1}', name=f'hyperparameter_{i}', type='float')
                            for i in range(self.num_hyperparameters)]
        hyper_parameters_bounds = [SimpleParameter(title=f'Hyperparameter #{i + 1} {edge}', name=f'hyperparameter_{i}_{edge}', type='float')
                                   for i in range(self.num_hyperparameters) for edge in ['min', 'max']]
        bounds_parameters = [SimpleParameter(title=f'Axis #{i + 1} {edge}', name=f'axis_{i}_{edge}', type='float')
                             for i in range(self.dimensionality) for edge in ['min', 'max']]
        func_parameters = [ListParameter(title='Method', name='method', limits=['global', 'local', 'hgdl'], default='global'),
                           ListParameter(title='Acquisition Function', name='acquisition_function', limits=list(gpcam_acquisition_functions.keys()), default=list(gpcam_acquisition_functions.keys())[0]),
                           SimpleParameter(title='Queue Length', name='n', value=1, type='int'),
                           SimpleParameter(title='Population Size (global only)', name='pop_size', value=20, type='int'),
                           SimpleParameter(title='Tolerance', name='tol', value=1e-6, type='float')]

        global_train_parameter = TrainingParameter(title='Train globally at...', name='global_training', addText='Add', children=[
            SimpleParameter(title='N=', name=str(uuid.uuid4()), value=N, type='int') for N in self.default_retrain_globally_at
        ])
        local_train_parameter = TrainingParameter(title='Train locally at...', name='local_training', addText='Add', children=[
            SimpleParameter(title='N=', name=str(uuid.uuid4()), value=N, type='int') for N in self.default_retrain_locally_at
        ])
        mcmc_train_parameter = TrainingParameter(title='Train locally at...', name='mcmc_training', addText='Add',
                                                  children=[
                                                      SimpleParameter(title='N=', name=str(uuid.uuid4()), value=N,
                                                                      type='int') for N in
                                                      self.default_retrain_mcmc_at
                                                  ])

        # wireup callback-based parameters
        for param in hyper_parameters:
            param.sigValueChanged.connect(self._set_hyperparameter)

        parameters = func_parameters + [GroupParameter(name='bounds', title='Axis Bounds', children=bounds_parameters),
                                        GroupParameter(name='hyperparameters', title='Hyperparameter Bounds', children=hyper_parameters + hyper_parameters_bounds),
                                        global_train_parameter,
                                        local_train_parameter,
                                        mcmc_train_parameter,
                                        ]
        return GroupParameter(name='top', children=parameters)

    def _set_hyperparameter(self, parameter, value):
        hyperparameters = np.asarray([self.parameters[('hyperparameters', f'hyperparameter_{i}')]
                                           for i in range(self.num_hyperparameters)])
        self.optimizer.set_hyperparameters(hyperparameters)

    def update_measurements(self, data: Data):
        with data.r_lock():  # quickly grab values within lock before passing to optimizer
            positions = data.positions.copy()
            scores = data.scores.copy()
            variances = data.variances.copy()
        self.optimizer.tell(np.asarray(positions), np.asarray(scores), np.asarray(variances))

    def init_gp(self, hyperparameters, **opts):
        self.optimizer.init_gp(hyperparameters, **opts)

    def update_metrics(self, data: Data):
        for graph in self.graphs:
            try:
                graph.compute(data, self)
            except Exception as ex:
                logger.exception(ex)

    def request_targets(self, position, **kwargs):
        self.last_position = position
        bounds = np.asarray([[self.parameters[('bounds', f'axis_{i}_{edge}')]
                     for edge in ['min', 'max']]
                    for i in range(self.dimensionality)])
        n = self.parameters['n']

        # If the GP is not initialized, generate random targets
        if not self.optimizer.gp:
            return [[np.random.uniform(bounds[i][0], bounds[i][1]) for i in range(self.dimensionality)] for _ in range(n)]
        else:
            kwargs.update({key: self.parameters[key] for key in ['acquisition_function', 'method', 'pop_size', 'tol']})
            kwargs.update({'input_set': bounds})
            kwargs.update(self.ask_opts)
            return self.optimizer.ask(position=position,
                                      n=n,
                                      acquisition_function=gpcam_acquisition_functions[kwargs.pop('acquisition_function')],
                                      x0=np.asarray(position),
                                      **kwargs)['x'].astype(float)

    def train(self):
        for method in ['global', 'local', 'mcmc']:
            train_at = set(child.value() for child in self.parameters.child(f'{method}_training').children())

            for N in train_at:
                if len(self.optimizer.y_data) > N and N not in self._completed_training[method]:
                    logger.info('Training in progress. This make take a while...')
                    self.optimizer.train(hyperparameter_bounds=
                                         np.asarray([[self.parameters[('hyperparameters', f'hyperparameter_{i}_{edge}')]
                                                      for edge in ['min', 'max']]
                                                     for i in range(self.num_hyperparameters)]),
                                         init_hyperparameters=
                                         np.asarray([self.parameters[('hyperparameters', f'hyperparameter_{i}')]
                                                     for i in range(self.num_hyperparameters)]), method=method)
                    self._completed_training[method].add(N)
                    logger.info(f"New hyperparameters: {self.optimizer.get_hyperparameters()}")

        return True



import logging
import time
from functools import partial

import numpy as np
import qdarkstyle
from PySide2.QtWidgets import QFormLayout, QPushButton, QSpacerItem, QLabel, QSizePolicy
# from caproto.sync.client import write, read
from bluesky.plan_stubs import mov, checkpoint, stage, sleep
from bluesky.preprocessors import run_decorator
from gpcam.autonomous_experimenter import AutonomousExperimenterGP
from ophyd import EpicsSignalRO, EpicsMotor
from qtpy.QtWidgets import QApplication, QWidget, QHBoxLayout

from tsuchinoko.plan_stubs import tune_max_and_fit
from tsuchinoko.utils import runengine
from tsuchinoko.utils.threads import invoke_in_main_thread
# TODO: goal values: FWHM 5 microns, 50 nA
# TODO: backlash: none
# TODO: positioner precision:
# TODO: photodiode accuracy: .0005 nA
# TODO: ring mode: 2-bunch mode has stable current
# TODO: measurement time: integrated over .5s
# TODO: mirror movement speed: fast
# TODO: device names, units: diode in nA
# TODO: limits
# TODO: test device control
# TODO:
# BCS701:DetectorDiodeCurrent:ai-AI
# BCS701:PinholeX; microns ; +- 14500 Y; +-6000 X; accurate to 10 nm
# BCS701:PinholeY
# BCS701:M112HorizAngle
# BCS701:M112VertAngle
from tsuchinoko.widgets.displays import RunEngineControls

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)s] %(message)s",
    handlers=[
        # logging.FileHandler("debug.log"),
        logging.StreamHandler()
    ]
)

PINHOLE_TRACK_FACTOR_X = 199 / .05 / .9  # micron per mRad
PINHOLE_TRACK_FACTOR_Y = 132  # micron per mRad
DARK_CURRENT = .035


def acq_func(x, gp):
    m = gp.posterior_mean(x)["f(x)"]
    v = gp.posterior_covariance(x)["v(x)"]
    # print('val:', m, v)
    return m + (posterior_weight_factor.value()) * np.sqrt(v)


def measure_quality(data, monitor, motor1, motor2, pinhole1, pinhole1_scan_range, pinhole2, pinhole2_scan_range,
                    pinhole_min_step):
    for entry in data:
        cycle_start = time.time()
        last_motor_position = motor1.user_readback.get(), motor2.user_readback.get()
        last_pinhole_position = pinhole1.user_readback.get(), pinhole2.user_readback.get()
        next_motor_position = entry['position']
        pinhole_relative_position = (next_motor_position[0] - last_motor_position[0]) * PINHOLE_TRACK_FACTOR_X, (next_motor_position[1] - last_motor_position[1]) * PINHOLE_TRACK_FACTOR_Y
        next_pinhole_position = pinhole_relative_position[0] + last_pinhole_position[0], pinhole_relative_position[1] + last_pinhole_position[1]
        # move to position
        logging.info(msg=f'Moving to position: {next_motor_position}')
        logging.info(msg=f'Moving pinhole to track beam: {next_pinhole_position}')
        yield from mov(motor1, next_motor_position[0], motor2, next_motor_position[1], pinhole1, next_pinhole_position[0], pinhole2, next_pinhole_position[1])
        yield from sleep(.5)

        # move pinhole to centroid and fit along the way
        logging.info(msg=f'Looking for centroid with pinhole...')
        # x_fit = (yield from tune_centroid_and_fit([monitor], 'monitor', pinhole1, pinhole1min, pinhole1max, pinhole_min_step, snake=True, name='x_fit', expected_spot_size=1))
        x_fit = (yield from tune_max_and_fit([monitor], 'monitor', pinhole1, next_pinhole_position[0] - pinhole1_scan_range / 2, next_pinhole_position[0] + pinhole1_scan_range / 2, pinhole_min_step, snake=True, name='x_fit', expected_spot_size=40, debug=configuration.debug_fit.isChecked(), sleep=.5, dark_value=DARK_CURRENT))
        y_fit = (yield from tune_max_and_fit([monitor], 'monitor', pinhole2, next_pinhole_position[1] - pinhole2_scan_range / 2, next_pinhole_position[1] + pinhole2_scan_range / 2, pinhole_min_step, snake=True, name='y_fit', expected_spot_size=40, debug=configuration.debug_fit.isChecked(), sleep=.5, dark_value=DARK_CURRENT))
        if x_fit and y_fit:

            logging.info(msg=f'Max estimated at: {x_fit.mean_0.value}, {y_fit.mean_0.value}')
            logging.info(msg=f'X fit:\n    amplitude: {x_fit.amplitude_0.value:.3f}\n    stddev: {x_fit.stddev_0.value:.3f}')
            logging.info(msg=f'Y fit:\n    amplitude: {y_fit.amplitude_0.value:.3f}\n    stddev: {y_fit.stddev_0.value:.3f}')

            # TODO: check error between x_amp, y_amp, and measured amp

            # check divergence from tracking; ideally this would be 1, 1 if the TRACK_FACTORs are correct
            tracking_factor_error_x = pinhole_relative_position[0] / (x_fit.mean_0.value - last_pinhole_position[0])
            tracking_factor_error_y = pinhole_relative_position[1] / (y_fit.mean_0.value - last_pinhole_position[1])

            logging.info(msg=f'Tracking factor error: {tracking_factor_error_x}, {tracking_factor_error_y}')
            if abs(tracking_factor_error_x - 1) > .1:
                logging.warning(f'{motor1.name} is not tracking well; moved by {pinhole_relative_position[0]} when should have moved by {x_fit.mean_0.value - last_pinhole_position[0]}')
            if abs(tracking_factor_error_y - 1) > .1:
                logging.warning(f'{motor2.name} is not tracking well; moved by {pinhole_relative_position[1]} when should have moved by {y_fit.mean_0.value - last_pinhole_position[1]}')

            # measure

            # get value
            metric_factors = [configuration.mean_weight.value(), configuration.stddev_x_weight.value(), configuration.stddev_y_weight.value()]
            metric_vec = np.asarray([y_fit.amplitude_0.value, 1 / x_fit.stddev_0.value, 1 / y_fit.stddev_0.value]) * np.asarray(metric_factors)
            logging.info(msg=f'metrics: {metric_vec}')
            entry['value'] = np.linalg.norm(metric_vec)
            entry['variance'] = measure_variance
        else:
            entry['value'] = 0
        RunEngineControls().measurement_time.setText(f'{time.time() - cycle_start:.1f} s')
    return data


@run_decorator(md={})
def alignment_plan(monitor, motor1, min1, max1, motor2, min2, max2, pinhole1, pinhole1_scan_range, pinhole2,
                   pinhole2_scan_range, pinhole_min_step, N, **kwargs):
    # set up parameter space
    parameters = np.array([[min1, max1],
                           [min2, max2],
                           ])

    ##set up some hyperparameters, if you have no idea, set them to 1 and make the training bounds large
    init_hyperparameters = np.array([1, 1, 1])
    hyperparameter_bounds = np.array([[.1, 1],  # bounds square of 'value' range
                                      [0.01, 10],
                                      [0.01, 10]])

    delayed_data = None

    def delay_measure(data):
        global delayed_data
        delayed_data = data
        # for entry in data:
        #     entry['value'] = 0
        return data

    ##let's initialize the autonomous experimenter ...
    experiment = AutonomousExperimenterGP(parameters,
                                          delay_measure,
                                          init_hyperparameters,
                                          hyperparameter_bounds,
                                          init_dataset_size=10,
                                          # acq_func='ucb',
                                          acq_func=acq_func,
                                          run_every_iteration=show_result,
                                          append_data_after_send=True)
    experiment.instrument_func = partial(measure_quality,
                                         monitor=monitor,
                                         motor1=motor1,
                                         motor2=motor2,
                                         pinhole1=pinhole1,
                                         pinhole1_scan_range=pinhole1_scan_range,
                                         pinhole2=pinhole2,
                                         pinhole2_scan_range=pinhole2_scan_range,
                                         pinhole_min_step=pinhole_min_step)
    devices = [monitor, motor1, motor2, pinhole1, pinhole2]
    for device in devices:
        yield from stage(device)

    experiment.data.dataset = (yield from experiment.instrument_func(experiment.data.dataset))

    # TODO: handle yielding of init dataset actions

    # ...train...
    experiment.train()

    n = 0
    while n < N:
        cycle_start = time.time()
        res = experiment.gp_optimizer.ask(
            n=1,
            acquisition_function=experiment.acq_func,
            cost_function=experiment.cost_func,
            dask_client=experiment.acq_func_opt_dask_client)
        next_measurement_points = res["x"]
        post_var = experiment.gp_optimizer.posterior_covariance(next_measurement_points)["v(x)"]

        info = [{"hyperparameters": experiment.gp_optimizer.hyperparameters,
                 "posterior std": np.sqrt(post_var[j])} for j in range(len(next_measurement_points))]
        new_data = experiment.data.inject_arrays(next_measurement_points, info=info)
        new_data = yield from experiment.instrument_func(new_data)
        experiment.data.dataset = experiment.data.dataset + new_data
        experiment.x, experiment.y, experiment.v, experiment.t, experiment.c, vp = experiment.extract_data()
        experiment.tell(experiment.x, experiment.y, experiment.v, vp)
        if (len(experiment.data.dataset) % 10) == 0:
            experiment.train(pop_size=10, tol=1e-6, max_iter=20, method="global")
        invoke_in_main_thread(show_result, experiment)
        # show_result(experiment)
        yield from checkpoint()
        RunEngineControls().cycle_time.setText(f'{time.time() - cycle_start:.1f} s')

        n += 1


def show_result(experiment):
    global scatter
    x = [acq['position'][0] for acq in experiment.data.dataset]
    y = [acq['position'][1] for acq in experiment.data.dataset]
    v = [acq['value'] for acq in experiment.data.dataset]
    c = [255 * i / len(experiment.data.dataset) for i in range(len(experiment.data.dataset))]
    max_index = np.argmax(v)
    scatter.setData(
        [{'pos': (xi, yi),
          'size': vi / max(v) * 20 + 5,
          'brush': pg.mkBrush(color=pg.mkColor(255, 255, 255)) if i == len(x) - 1 else pg.mkBrush(
              color=pg.mkColor(255 - c, c, 0)),
          'symbol': '+' if i == len(x) - 1 else 'o'}
         for i, (xi, yi, vi, c) in enumerate(zip(x, y, v, c))])
    arrow.setIndex(max_index)
    text.setText(f'Max: ({x[max_index]:.2f}, {y[max_index]:.2f})')
    text.setPos(x[max_index], y[max_index])


if __name__ == '__main__':
    import pyqtgraph as pg

    # min1 = -3
    # max1 = 3
    # min2 = -3
    # max2 = 3
    # pinhole1min = -3
    # pinhole1max = 3
    # pinhole2min = -3
    # pinhole2max = 3
    # pinhole_min_step = .01

    min1 = -8 - .3  # hard limit -13.9
    max1 = -8 + .4  # hard limit 1.9
    min2 = -9.5 - .3  # hard limit -17.4
    max2 = -9.5 + .3  # hard limit 2.2
    # pinhole1min = -1600
    # pinhole1max = 1600
    # pinhole2min = 1850+200
    # pinhole2max = 1850-200
    pinhole1_scan_range = 500
    pinhole2_scan_range = 500
    pinhole_min_step = 5  # 10% of reasonable FWHM
    mean_weight_default, stddev_x_weight_default, stddev_y_weight_default = [1e0, 1e2, 1e2]
    posterior_weight_factor_default = 3
    measure_variance = 0.0005 ** 2  # TODO: Is this a unitless or unit-full value? a percentage?

    motor1 = EpicsMotor(prefix='BCS701:M112HorizAngle', name='motor1')  # initial -7.99216
    motor2 = EpicsMotor(prefix='BCS701:M112VertAngle', name='motor2')  # initial -5.49572
    pinhole1 = EpicsMotor(prefix='BCS701:PinholeX', name='pinhole1')
    pinhole2 = EpicsMotor(prefix='BCS701:PinholeY', name='pinhole2')

    monitor = EpicsSignalRO(name='monitor', read_pv='BCS701:DetectorDiodeCurrent:ai-AI')

    qapp = QApplication([])
    dark_stylesheet = qdarkstyle.load_stylesheet()
    qapp.setStyleSheet(dark_stylesheet)

    w = QWidget()
    outer_layout = QHBoxLayout()
    w.setLayout(outer_layout)
    outer_layout.setContentsMargins(0, 0, 0, 0)
    iv = pg.PlotWidget()
    scatter = pg.ScatterPlotItem(x=[0], y=[0], size=10, pen=pg.mkPen(None), brush=pg.mkBrush(255, 255, 255, 120))
    arrow = pg.CurveArrow(scatter)
    text = pg.TextItem()

    start = QPushButton('Start')
    pause = QPushButton('Pause')
    resume = QPushButton('Resume')
    pause.hide()
    resume.hide()
    measurement_time = QLabel('...')
    cycle_time = QLabel('...')

    w.layout().addWidget(iv)
    form_layout = QFormLayout()
    form_layout.addRow('Post. Mean/Covariance Weighting', posterior_weight_factor)
    form_layout.addRow('Amplitude weight', mean_weight)
    form_layout.addRow('Std. Dev. X weight', stddev_x_weight)
    form_layout.addRow('Std. Dev. Y weight', stddev_y_weight)
    form_layout.addRow('Debug centroid/tune/fit', debug_fit)
    form_layout.addWidget(start)
    form_layout.addWidget(pause)
    form_layout.addWidget(resume)
    form_layout.addItem(QSpacerItem(1, 1, vData=QSizePolicy.Expanding))
    form_layout.addRow('Measurement Time:', measurement_time)
    form_layout.addRow('Cycle Time:', cycle_time)
    w.layout().addLayout(form_layout)

    iv.addItem(scatter)
    iv.addItem(arrow)
    iv.addItem(text)
    w.show()
    N = 1000

    RE = runengine.get_run_engine()

    plan = alignment_plan(monitor, motor1, min1, max1, motor2, min2, max2, pinhole1, pinhole1_scan_range, pinhole2,
                          pinhole2_scan_range, pinhole_min_step, 10000)


    def start_plan():
        RE(plan)
        start.hide()
        pause.show()


    def pause_plan():
        RE.pause()
        resume.show()
        pause.hide()


    def resume_plan():
        RE.resume()
        resume.hide()
        pause.show()


    start.clicked.connect(start_plan)
    pause.clicked.connect(pause_plan)
    resume.clicked.connect(resume_plan)

    qapp.exec_()



import uuid

from pyqtgraph.parametertree import registerParameterType
from pyqtgraph.parametertree.parameterTypes import GroupParameter


class TrainingParameter(GroupParameter):
    def __init__(self, **kwargs):
        kwargs['type'] = 'training'
        super(TrainingParameter, self).__init__(**kwargs)

    def addNew(self):
        self.addChild(dict(name=str(uuid.uuid4()), title='N=', type='int', value=1, removable=True, renamable=False))


registerParameterType('training', TrainingParameter)



from collections import defaultdict
from contextlib import contextmanager
from copy import copy
from dataclasses import dataclass, field, asdict
from abc import ABC, abstractmethod
from typing import Tuple, Iterable, Set, List, Union

from loguru import logger
from pyqtgraph.parametertree import Parameter

from tsuchinoko.graphs import Graph
from tsuchinoko.utils.mutex import RWLock


@dataclass
class Data:
    """
    A data class to track the data state of an experiment. This type can be appended to as new data is received. A
    locking mechanism is provided to support read/write locking for parallelism.
    """

    dimensionality: int = None
    positions: list = field(default_factory=list)
    scores: list = field(default_factory=list)
    variances: list = field(default_factory=list)
    metrics: dict = field(default_factory=lambda: defaultdict(list))
    states: dict = field(default_factory=dict)
    graphics_items: dict = field(default_factory=dict)

    @property
    def measurements(self):
        return list(zip(self.positions, self.scores, self.variances, [{key: values[i] for key, values in self.metrics.items()} for i in range(len(self))]))

    def __post_init__(self):
        self._lock = RWLock()
        self.w_lock = self._lock.w_locked
        self.r_lock = self._lock.r_locked
        self._completed_iterations = 0

    def inject_new(self, data):
        with self.w_lock():
            for datum in data:
                self.positions.append(datum[0])
                self.scores.append(datum[1])
                self.variances.append(datum[2])
                for metric in datum[3]:  # TODO: handle logical cases
                    if metric not in self.metrics:
                        self.metrics[metric] = []
                    self.metrics[metric].append(datum[3][metric])

    def as_dict(self):
        self_copy = copy(self)
        self_copy.metrics = dict(self_copy.metrics)
        return asdict(self_copy)

    def __getitem__(self, item: Union[slice, str]):
        if isinstance(item, str):
            if item in self.metrics and item in self.states:
                raise ValueError(f'{item} exists in both states and metrics.')
            elif item in self.metrics:
                return self.metrics[item]
            elif item in self.states:
                return self.states[item]
            elif item.lower() in ['variances', 'scores', 'positions']:
                return getattr(self, item.lower())
        elif isinstance(item, slice):
            if item.stop is None or item.stop > len(self):
                item = slice(item.start, len(self), item.step)
            return Data(self.dimensionality,
                        self.positions[item],
                        self.scores[item],
                        self.variances[item],
                        {key: value[item] for key, value in self.metrics.items()},
                        self.states,
                        self.graphics_items)
        raise ValueError(f'Unknown item: {item}')

    def __contains__(self, item):
        if isinstance(item, str):
            return item in self.states or item in self.metrics

    def __setitem__(self, key, value):
        if isinstance(key, str):
            self.metrics[key] = value
        else:
            raise ValueError()

    def __len__(self):
        return len(self.positions)

    def extend(self, data: 'Data'):
        with self.w_lock():
            self.positions += data.positions
            self.scores += data.scores
            self.variances += data.variances
            for key in set(self.metrics) | set(data.metrics):
                self.metrics[key] += data.metrics.get(key, [])
            self.dimensionality = data.dimensionality
            self.graphics_items.update(data.graphics_items)
            self.states = data.states

    def __enter__(self):
        self.w_lock().__enter__()
        logger.exception(RuntimeError("deprecation in progress"))

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.w_lock().__exit__(exc_type, exc_val, exc_tb)

    def __bool__(self):
        return bool(len(self))

    @contextmanager
    def iteration(self):
        yield
        self._completed_iterations += 1


class Engine(ABC):
    """
    The Adaptive Engine base class. This component is generally to be responsible for determining future measurement targets.
    """

    dimensionality: int = None
    parameters: Parameter = None
    graphs: List['Graph'] = None
    last_position: tuple = None

    @abstractmethod
    def update_measurements(self, data: Data):
        """
        Update internal variables with the provided new data
        """
        ...

    @abstractmethod
    def request_targets(self, position: Tuple) -> Iterable[Tuple]:
        """
        Determine new targets to be measured

        Parameters
        ----------
        position: tuple
            The current 'position' of the experiment in the target domain.

        Returns
        -------
        targets: array_like
            The new targets to be measured.
        """
        ...

    @abstractmethod
    def reset(self):
        """
        Called when an experiment stops, or is about to start. Returns the engine to a clean state.
        """
        ...

    @abstractmethod
    def train(self):
        """
        Perform training. This can be short-circuited to only train on every N-th iteration, for example.
        """
        ...

    @abstractmethod
    def update_metrics(self, data: Data):
        """
        Calculates various metrics to drive visualizations for the client. The data object is expected to be mutated to
        include these new values.
        """
        ...



from threading import Thread

from pytest import fixture
from qtpy.QtWidgets import QMessageBox, QFileDialog
from pytestqt import qtbot
from qtpy import QtCore
from loguru import logger

from tsuchinoko.core import ConnectResponse, ZMQCore, CoreState
from tsuchinoko.widgets.mainwindow import MainWindow

# Disable logging to console when running tests
# NOTE: it seems there is a bug between loguru and pytest; pytest tries to log to a tempfile, but closes it when finished
# NOTE: if loguru has a backlog of messages
# logger.remove()


@fixture
def client_window(qtbot):
    logger.info('starting client window setup')
    main_window = MainWindow()
    main_window.show()
    logger.info('client window setup complete')
    # qtbot.addWidget(main_window)
    with qtbot.wait_exposed(main_window):
        yield main_window

    logger.info('teardown client window')
    with qtbot.waitCallback() as cb:
        main_window.update_thread.sigFinished.connect(cb)
        main_window.close()
    logger.info('client window teardown finished')


@fixture
def dialog_response_no(monkeypatch):
    # Suppress save dialog
    with monkeypatch.context() as m:
        m.setattr(QMessageBox, "question", lambda *args, **kwargs: QMessageBox.No)
        yield


@fixture
def client_and_server(qtbot, dialog_response_no, random_engine, simple_execution_engine, client_window):

    core = ZMQCore()
    core.set_execution_engine(simple_execution_engine)
    core.set_adaptive_engine(random_engine)

    with qtbot.waitCallback() as cb:
        client_window.subscribe(cb, ConnectResponse)
        server_thread = Thread(target=core.main)
        server_thread.start()
        if client_window.state_manager_widget.state != CoreState.Connecting:
            cb()

    def button_enabled():
        assert client_window.state_manager_widget.start_pause_button.isEnabled()

    qtbot.waitUntil(button_enabled)
    qtbot.mouseClick(client_window.state_manager_widget.start_pause_button, QtCore.Qt.LeftButton)
    qtbot.waitUntil(lambda: len(client_window.data) > 0)

    yield client_window, core

    qtbot.mouseClick(client_window.state_manager_widget.stop_button, QtCore.Qt.LeftButton)

    qtbot.waitUntil(lambda: client_window.state_manager_widget.state == CoreState.Inactive)

    core.exit()
    server_thread.join()


def test_simple(client_and_server):
    client_window, core = client_and_server
    assert len(client_window.data) > 0


def test_save_restore(client_and_server, monkeypatch, tmp_path):
    client_window, core = client_and_server

    with monkeypatch.context() as m:
        m.setattr(QFileDialog, 'getSaveFileName', lambda filter: (tmp_path / 'test.yml', "YAML (*.yml)"))
        client_window.save_data()

    with monkeypatch.context() as m:
        m.setattr(QMessageBox, 'question', lambda *args, **kwargs: QMessageBox.Yes)
        client_window.new_data()

    with monkeypatch.context() as m:
        m.setattr(QFileDialog, 'getOpenFileName', lambda filter: (tmp_path / 'test.yml', "YAML (*.yml)"))
        m.setattr(QMessageBox, 'question', lambda *args, **kwargs: QMessageBox.Yes)
        client_window.open_data()

    with monkeypatch.context() as m:
        m.setattr(QFileDialog, 'getSaveFileName', lambda filter: (tmp_path / 'test_params.yml', "YAML (*.yml)"))
        client_window.save_parameters()

    with monkeypatch.context() as m:
        m.setattr(QFileDialog, 'getOpenFileName', lambda filter: (tmp_path / 'test_params.yml', "YAML (*.yml)"))
        client_window.open_parameters()



import numpy as np
from ophyd.sim import SynAxis, Device, Cpt, SynSignalRO, SynSignal
from bluesky.plan_stubs import mv, trigger_and_read, create, stage, checkpoint, mov
from PIL import Image

from tsuchinoko.adaptive.gpCAM_in_process import GPCAMInProcessEngine
from tsuchinoko.core import Core
from tsuchinoko.execution.bluesky_in_process import BlueskyInProcessEngine


if __name__ == '__main__':

    image = np.asarray(Image.open('example_assets/sombrero_pug.jpg'))

    class PointDetector(Device):
        motor1 = Cpt(SynAxis, name='motor1')
        motor2 = Cpt(SynAxis, name='motor2')
        value = Cpt(SynSignal, name='value')

        def __init__(self, name):

            super(PointDetector, self).__init__(name=name)

            self.value.sim_set_func(self.get_value)

        def get_value(self):
            return np.average(image[int(self.motor1.position), int(self.motor2.position)])

        def trigger(self, *args, **kwargs):
            return self.value.trigger(*args, **kwargs)


    point_detector = PointDetector('point_detector')

    def measure_target(target):
        yield from checkpoint()
        yield from mov(point_detector.motor1, target[0], point_detector.motor2, target[1])
        ret = (yield from trigger_and_read([point_detector]))
        return ret[point_detector.value.name]['value']

    def get_position():
        yield from checkpoint()
        return point_detector.motor1.position, point_detector.motor2.position


    adaptive = GPCAMInProcessEngine(dimensionality=2,
                                    parameter_bounds=[(0, image.shape[0]),
                                                      (0, image.shape[1])],
                                    hyperparameters=[1, 1, 1],
                                    hyperparameter_bounds=[(0, 255),
                                                           (0, 1e3),
                                                           (0, 1e3)])
    execution = BlueskyInProcessEngine(measure_target, get_position)

    core = Core()
    core.set_adaptive_engine(adaptive)
    core.set_execution_engine(execution)

    core.start()




import numpy as np
from tsuchinoko.adaptive.adaptive import Adaptive2D
from tsuchinoko.execution.simple import SimplestEngine
from tsuchinoko.core import ZMQCore
import time


def my_measurement_function(position, fwhm=.5):  # (a gaussian)
    time.sleep(.1)
    return np.exp(-4 * np.log(2) * ((position[0]) ** 2 + (position[1]) ** 2) / fwhm ** 2)


if __name__ == '__main__':
    # Define an execution engine
    execution = SimplestEngine(measure_func=my_measurement_function)

    # Define an adaptive engine with initial parameters
    adaptive = Adaptive2D(parameter_bounds=[(-1, 1), (-1, 1)])

    # Start the core server
    ZMQCore(adaptive_engine=adaptive, execution_engine=execution).main()

# Now open a Tsuchinoko client or launch from Server Editor to watch the experiment
# $> tsuchinoko



from concurrent.futures import ProcessPoolExecutor
from functools import cached_property
# from queue import Queue

import adaptive
from . import _adaptive
from pyqtgraph.parametertree.parameterTypes import SimpleParameter, GroupParameter

from tsuchinoko.adaptive import Engine, Data
from tsuchinoko.graphs.common import Variance, Score
from tsuchinoko.utils import threads





class Adaptive2D(Engine):
    dimensionality: int = 2

    def __init__(self, parameter_bounds=None):
        for i in range(self.dimensionality):
            for j, edge in enumerate(['min', 'max']):
                self.parameters[('bounds', f'axis_{i}_{edge}')] = parameter_bounds[i][j]

        self.graphs = [Variance(), Score(), ]

        self.executor = None
        self.thread = None

        self.reset()

    # def __getstate__(self):
    #     state = self.__dict__.copy()
    #     del state['parameters']
    #     del state['thread']
    #     del state['learner']
    #     del state['runner']
    #
    #     return state

    @cached_property
    def parameters(self):
        bounds_parameters = [SimpleParameter(title=f'Axis #{i + 1} {edge}', name=f'axis_{i}_{edge}', type='float')
                             for i in range(self.dimensionality) for edge in ['min', 'max']]

        parameters = [GroupParameter(name='bounds', title='Axis Bounds', children=bounds_parameters), ]
        return GroupParameter(name='top', children=parameters)

    def update_measurements(self, data: Data):
        _adaptive.measurements.put(data.scores[-1])

    def request_targets(self, position, **kwargs):
        if not self.thread.isRunning():
            self.thread.start()
        return [_adaptive.targets.get()]

    def reset(self):
        # TODO: build safe exit for runner thread

        bounds = [tuple(self.parameters[('bounds', f'axis_{i}_{edge}')]
                  for edge in ['min', 'max'])
                  for i in range(self.dimensionality)]

        # clear queues
        while not _adaptive.targets.empty():
            _adaptive.targets.get()
        while not _adaptive.measurements.empty():
            _adaptive.measurements.get()

        self.learner = adaptive.Learner2D(_adaptive.measure, bounds=bounds)
        self.executor = ProcessPoolExecutor(initializer=_adaptive.set_global, initargs=(_adaptive.targets, _adaptive.measurements))

        # self.runner = adaptive.Runner(self.learner, , goal=lambda l: False)
        self.thread = threads.QThreadFuture(adaptive.BlockingRunner, self.learner, executor=self.executor,
                                            goal=lambda l: False)

    def train(self):
        ...

    def update_metrics(self, data: Data):
        ...



from pathlib import Path
import numpy as np
from PIL import Image
from pyqtgraph import mkColor
from scipy import ndimage
from itertools import count

from tsuchinoko.adaptive.gpCAM_in_process import GPCAMInProcessEngine
from tsuchinoko.core import ZMQCore
from tsuchinoko.execution.simple import SimpleEngine
from tsuchinoko.graphs.common import Plot, MultiPlot, DynamicColorMultiPlot

# Load data from a jpg image to be used as a luminosity map
image = np.flipud(np.asarray(Image.open(Path(__file__).parent/'sombrero_pug.jpg')))
luminosity = np.average(image, axis=2)

# A counter to track measurement number
counter = count(1)

# rosenbrock?
# "pos" is 'annealing curve'
# scalar_color is 'grain size' and derived from ackley function of curve values

def objective(x, y):
    return -20.0 * np.exp(-0.2 * np.sqrt(0.5 * (x ** 2 + y ** 2))) - np.exp(0.5 * (np.cos(2 * np.pi * x) + np.cos(2 * np.pi * y))) + np.e + 20


# Bilinear sampling will be used to effectively smooth pixel edges in source data
def bilinear_sample(pos):
    i = next(counter)
    v = objective(pos[0], np.arange(-20, 20))
    grain_size = np.random.random()
    return pos, \
        ndimage.map_coordinates(luminosity, [[pos[1]], [pos[0]]], order=1)[0], \
        1, \
        {'plot_data': v,
         'plot_label': f'Measurement #{i}',
         # 'pen': {'width': int(np.max(v)), 'color': np.min(v)}
         'scalar_color': grain_size
         }


execution = SimpleEngine(measure_func=bilinear_sample)


# Define a gpCAM adaptive engine with initial parameters
adaptive = GPCAMInProcessEngine(dimensionality=2,
                                parameter_bounds=[(0, image.shape[1]),
                                                  (0, image.shape[0])],
                                hyperparameters=[255, 100, 100],
                                hyperparameter_bounds=[(0, 1e5),
                                                       (0, 1e5),
                                                       (0, 1e5)])

multiplot_graph_example = DynamicColorMultiPlot(color_scalar_key='scalar_color', data_key='plot_data', name='Vector Metric Example', label_key='plot_label', accumulates=True)

adaptive.graphs.append(multiplot_graph_example)

# Construct a core server
core = ZMQCore()
core.set_adaptive_engine(adaptive)
core.set_execution_engine(execution)

if __name__ == '__main__':
    # Start the core server
    core.main()



import enum
import struct

from pyqode.core.api.client import JsonTcpClient, comm  # Must be a late import
from qtpy import QtGui, QtCore, QtWidgets


def _read_header(self):
    comm('reading header')
    self._header_buf += self.read(4)
    if len(self._header_buf) == 4:
        self._header_complete = True
        try:
            if hasattr(self._header_buf, 'data'):
                raise TypeError  # The following line unforgivingly causes access violation on PySide2, skip to doing it right
            header = struct.unpack('=I', self._header_buf)
        except TypeError:
            # pyside
            header = struct.unpack('=I', self._header_buf.data())
        self._to_read = header[0]
        self._header_buf = bytes()
        comm('header content: %d', self._to_read)


JsonTcpClient._read_header = _read_header

# fix enum promotion in qtpy/pyside6. This looks wacky, but it actually does something.
for module in [QtCore, QtWidgets, QtGui]:
    class_names = [name for name in dir(module) if name.startswith("Q")]
    for class_name in class_names:
        klass = getattr(module, class_name)
        attrib_names = [name for name in dir(klass) if name[0].isupper()]
        for attrib_name in attrib_names:
            attrib = getattr(klass, attrib_name)
            if not isinstance(attrib, enum.EnumMeta):
                continue
            for name, value in attrib.__members__.items():
                setattr(klass, name, value)



from pathlib import Path
import numpy as np
from PIL import Image
from scipy import ndimage
import numpy as np

from tsuchinoko.adaptive.quadtree import QuadTreeEngine
from tsuchinoko.core import ZMQCore
from tsuchinoko.execution.simple import SimpleEngine

# Load data from a jpg image to be used as a luminosity map
image = np.flipud(np.asarray(Image.open(Path(__file__).parent / 'sombrero_pug.jpg')))
luminosity = np.average(image, axis=2)


# Bilinear sampling will be used to effectively smooth pixel edges in source data
def bilinear_sample(pos):
    return pos, ndimage.map_coordinates(luminosity, [[pos[1]], [pos[0]]], order=1)[0], 1, {}


# Poisson noise applied to measured value; variance is set in accordance with poisson statistics
def noisy_sample(pos):
    pos, value, variance, extra = bilinear_sample(pos)
    return pos, np.random.poisson(value), value, extra


def gaussian_sample(pos):
    pos, value, variance, extra = bilinear_sample(pos)
    return pos, np.random.standard_normal(1)[0] * 30 + value, value, extra


execution = SimpleEngine(measure_func=bilinear_sample)

# Define a gpCAM adaptive engine with initial parameters
adaptive = QuadTreeEngine(
    parameter_bounds=[(0, image.shape[1]),
                      (0, image.shape[0])],
    hyperparameters=[255, 100, 100],
    hyperparameter_bounds=[(0, 1e5),
                           (0, 1e5),
                           (0, 1e5)]
)

# Construct a core server
core = ZMQCore()
core.set_adaptive_engine(adaptive)
core.set_execution_engine(execution)

if __name__ == '__main__':
    # Start the core server
    core.main()



from pathlib import Path
import numpy as np
from PIL import Image
from bluesky.plan_stubs import trigger_and_read, checkpoint, mov
import ophyd  # insulates from ophyd import errors
from ophyd.sim import SynAxis, Device, Cpt, SynSignal
from scipy import ndimage

from tsuchinoko.adaptive.gpCAM_in_process import GPCAMInProcessEngine
from tsuchinoko.core import ZMQCore
from tsuchinoko.execution.bluesky_in_process import BlueskyInProcessEngine

# Load data from a jpg image to be used as a luminosity map
image = np.flipud(np.asarray(Image.open(Path(__file__).parent/'sombrero_pug.jpg')))
luminosity = np.average(image, axis=2)

# Bilinear sampling will be used to effectively smooth pixel edges in source data
def bilinear_sample(img, pos):
    return ndimage.map_coordinates(img, [[pos[0]], [pos[1]]], order=1)

# Define a simulated Ophyd Device class for measurement and control of position;
# this pulls measurements from the image data above
class PointDetector(Device):
    motor1 = Cpt(SynAxis, name='motor1')
    motor2 = Cpt(SynAxis, name='motor2')
    value = Cpt(SynSignal, name='value')

    def __init__(self, name):
        super(PointDetector, self).__init__(name=name)
        self.value.sim_set_func(self.get_value)

    def get_value(self):
        return np.average(bilinear_sample(luminosity, [int(self.motor2.position), int(self.motor1.position)]))

    def trigger(self, *args, **kwargs):
        return self.value.trigger(*args, **kwargs)

# Instantiate a simulated device
point_detector = PointDetector('point_detector')

# Define a Bluesky Plan component for performing measurements at targets.
# Note that this returns the measured value and variance
def measure_target(target):
    yield from checkpoint()
    yield from mov(point_detector.motor1, target[0], point_detector.motor2, target[1])
    ret = (yield from trigger_and_read([point_detector]))
    return ret[point_detector.value.name]['value'], 2  # variance of 1

# Define a Bluesky Plan component to get the current position of the device
# (which may not necessarily be precisely the requested position)
def get_position():
    yield from checkpoint()
    return point_detector.motor1.position, point_detector.motor2.position

# Define an execution engine with the measurement and get_position functions
execution = BlueskyInProcessEngine(measure_target, get_position)

# Define a gpCAM adaptive engine with initial parameters
adaptive = GPCAMInProcessEngine(dimensionality=2,
                                parameter_bounds=[(0, image.shape[1]),
                                                  (0, image.shape[0])],
                                hyperparameters=[255, 100, 100],
                                hyperparameter_bounds=[(0, 1e5),
                                                       (0, 1e5),
                                                       (0, 1e5)])

# Construct a core server
core = ZMQCore()
core.set_adaptive_engine(adaptive)
core.set_execution_engine(execution)

if __name__ == '__main__':
    # Start the core server
    core.main()
